#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

typos = {
    'feature': 'features',
    'sources': 'source',
    'targets': 'target',
    'include': 'includes',
    'export_include': 'export_includes',
    'define': 'defines',
    'importpath': 'includes',
    'installpath': 'install_path',
    'iscopy': 'is_copy',
    'uses': 'use',
}
meths_typos = ['__call__', 'program', 'shlib', 'stlib', 'objects']
import sys
from waflib import Logs, Build, Node, Task, TaskGen, ConfigSet, Errors, Utils
from waflib.Tools import ccroot


def check_same_targets(self):
    mp = Utils.defaultdict(list)
    uids = {}

    def check_task(tsk):
        if not isinstance(tsk, Task.Task):
            return
        if hasattr(tsk, 'no_errcheck_out'):
            return
        for node in tsk.outputs:
            mp[node].append(tsk)
        try:
            uids[tsk.uid()].append(tsk)
        except KeyError:
            uids[tsk.uid()] = [tsk]

    for g in self.groups:
        for tg in g:
            try:
                for tsk in tg.tasks:
                    check_task(tsk)
            except AttributeError:
                check_task(tg)
    dupe = False
    for (k, v) in mp.items():
        if len(v) > 1:
            dupe = True
            msg = '* Node %r is created more than once%s. The task generators are:' % (
                k, Logs.verbose == 1 and " (full message on 'waf -v -v')" or ""
            )
            Logs.error(msg)
            for x in v:
                if Logs.verbose > 1:
                    Logs.error('  %d. %r', 1 + v.index(x), x.generator)
                else:
                    Logs.error('  %d. %r in %r', 1 + v.index(x), x.generator.name, getattr(x.generator, 'path', None))
            Logs.error('If you think that this is an error, set no_errcheck_out on the task instance')
    if not dupe:
        for (k, v) in uids.items():
            if len(v) > 1:
                Logs.error(
                    '* Several tasks use the same identifier. Please check the information on\n   https://waf.io/apidocs/Task.html?highlight=uid#waflib.Task.Task.uid'
                )
                tg_details = tsk.generator.name
                if Logs.verbose > 2:
                    tg_details = tsk.generator
                for tsk in v:
                    Logs.error('  - object %r (%r) defined in %r', tsk.__class__.__name__, tsk, tg_details)


def check_invalid_constraints(self):
    feat = set()
    for x in list(TaskGen.feats.values()):
        feat.union(set(x))
    for (x, y) in TaskGen.task_gen.prec.items():
        feat.add(x)
        feat.union(set(y))
    ext = set()
    for x in TaskGen.task_gen.mappings.values():
        ext.add(x.__name__)
    invalid = ext & feat
    if invalid:
        Logs.error(
            'The methods %r have invalid annotations:  @extension <-> @feature/@before_method/@after_method',
            list(invalid)
        )
    for cls in list(Task.classes.values()):
        if sys.hexversion > 0x3000000 and issubclass(cls, Task.Task) and isinstance(cls.hcode, str):
            raise Errors.WafError(
                'Class %r has hcode value %r of type <str>, expecting <bytes> (use Utils.h_cmd() ?)' % (cls, cls.hcode)
            )
        for x in ('before', 'after'):
            for y in Utils.to_list(getattr(cls, x, [])):
                if not Task.classes.get(y):
                    Logs.error('Erroneous order constraint %r=%r on task class %r', x, y, cls.__name__)
        if getattr(cls, 'rule', None):
            Logs.error('Erroneous attribute "rule" on task class %r (rename to "run_str")', cls.__name__)


def replace(m):
    oldcall = getattr(Build.BuildContext, m)

    def call(self, *k, **kw):
        ret = oldcall(self, *k, **kw)
        for x in typos:
            if x in kw:
                if x == 'iscopy' and 'subst' in getattr(self, 'features', ''):
                    continue
                Logs.error('Fix the typo %r -> %r on %r', x, typos[x], ret)
        return ret

    setattr(Build.BuildContext, m, call)


def enhance_lib():
    for m in meths_typos:
        replace(m)

    def ant_glob(self, *k, **kw):
        if k:
            lst = Utils.to_list(k[0])
            for pat in lst:
                sp = pat.split('/')
                if '..' in sp:
                    Logs.error("In ant_glob pattern %r: '..' means 'two dots', not 'parent directory'", k[0])
                if '.' in sp:
                    Logs.error("In ant_glob pattern %r: '.' means 'one dot', not 'current directory'", k[0])
        return self.old_ant_glob(*k, **kw)

    Node.Node.old_ant_glob = Node.Node.ant_glob
    Node.Node.ant_glob = ant_glob

    def ant_iter(self, accept=None, maxdepth=25, pats=[], dir=False, src=True, remove=True, quiet=False):
        if remove:
            try:
                if self.is_child_of(self.ctx.bldnode) and not quiet:
                    quiet = True
                    Logs.error(
                        'Calling ant_glob on build folders (%r) is dangerous: add quiet=True / remove=False', self
                    )
            except AttributeError:
                pass
        return self.old_ant_iter(accept, maxdepth, pats, dir, src, remove, quiet)

    Node.Node.old_ant_iter = Node.Node.ant_iter
    Node.Node.ant_iter = ant_iter
    old = Task.is_before

    def is_before(t1, t2):
        ret = old(t1, t2)
        if ret and old(t2, t1):
            Logs.error('Contradictory order constraints in classes %r %r', t1, t2)
        return ret

    Task.is_before = is_before

    def check_err_features(self):
        lst = self.to_list(self.features)
        if 'shlib' in lst:
            Logs.error('feature shlib -> cshlib, dshlib or cxxshlib')
        for x in ('c', 'cxx', 'd', 'fc'):
            if not x in lst and lst and lst[0] in [x + y for y in ('program', 'shlib', 'stlib')]:
                Logs.error('%r features is probably missing %r', self, x)

    TaskGen.feature('*')(check_err_features)

    def check_err_order(self):
        if not hasattr(self, 'rule') and not 'subst' in Utils.to_list(self.features):
            for x in ('before', 'after', 'ext_in', 'ext_out'):
                if hasattr(self, x):
                    Logs.warn('Erroneous order constraint %r on non-rule based task generator %r', x, self)
        else:
            for x in ('before', 'after'):
                for y in self.to_list(getattr(self, x, [])):
                    if not Task.classes.get(y):
                        Logs.error('Erroneous order constraint %s=%r on %r (no such class)', x, y, self)

    TaskGen.feature('*')(check_err_order)

    def check_compile(self):
        check_invalid_constraints(self)
        try:
            ret = self.orig_compile()
        finally:
            check_same_targets(self)
        return ret

    Build.BuildContext.orig_compile = Build.BuildContext.compile
    Build.BuildContext.compile = check_compile

    def use_rec(self, name, **kw):
        try:
            y = self.bld.get_tgen_by_name(name)
        except Errors.WafError:
            pass
        else:
            idx = self.bld.get_group_idx(self)
            odx = self.bld.get_group_idx(y)
            if odx > idx:
                msg = "Invalid 'use' across build groups:"
                if Logs.verbose > 1:
                    msg += '\n  target %r\n  uses:\n  %r' % (self, y)
                else:
                    msg += " %r uses %r (try 'waf -v -v' for the full error)" % (self.name, name)
                raise Errors.WafError(msg)
        self.orig_use_rec(name, **kw)

    TaskGen.task_gen.orig_use_rec = TaskGen.task_gen.use_rec
    TaskGen.task_gen.use_rec = use_rec

    def _getattr(self, name, default=None):
        if name == 'append' or name == 'add':
            raise Errors.WafError('env.append and env.add do not exist: use env.append_value/env.append_unique')
        elif name == 'prepend':
            raise Errors.WafError('env.prepend does not exist: use env.prepend_value')
        if name in self.__slots__:
            return super(ConfigSet.ConfigSet, self).__getattr__(name, default)
        else:
            return self[name]

    ConfigSet.ConfigSet.__getattr__ = _getattr


def options(opt):
    enhance_lib()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
import os
import shutil

import pytest

from PyInstaller import HOMEPATH, PLATFORM
from PyInstaller.utils.tests import importorskip

from PyInstaller.utils.win32 import winmanifest

#- Helpers


# Go from current DOM element to the top-level node, and collect element tags along the way. Then reverse the list so
# it goes from top-level element to the current one.
def _get_parent_tags(node):
    tags = []
    while node is not None:
        tags.append(node.tag)
        node = node.getparent()
    return list(reversed(tags))


# Remove the whitespace and newline change actions
def _filter_whitespace_changes(actions):
    import xmldiff.actions

    filtered_actions = []
    for action in actions:
        # Skip text updates that involve only whitespace. This involves both trailing text and text within the nodes.
        if isinstance(action, (xmldiff.actions.UpdateTextAfter, xmldiff.actions.UpdateTextIn)):
            if not action.text.strip():
                continue
        filtered_actions.append(action)
    return filtered_actions


# List of element tags from top-level <assembly> to <requestedExecutionLevel> for UAC settings.
_REQUEST_EXECUTION_LEVEL_TAGS = [
    '{urn:schemas-microsoft-com:asm.v1}assembly',
    '{urn:schemas-microsoft-com:asm.v3}trustInfo',
    '{urn:schemas-microsoft-com:asm.v3}security',
    '{urn:schemas-microsoft-com:asm.v3}requestedPrivileges',
    '{urn:schemas-microsoft-com:asm.v3}requestedExecutionLevel',
]

# List of element tags from top-level <assembly> to <assemblyIdentity> for dependent assemblies.
_DEPENDENT_ASSEMBLY_IDENTITY_TAGS = [
    '{urn:schemas-microsoft-com:asm.v1}assembly',
    '{urn:schemas-microsoft-com:asm.v1}dependency',
    '{urn:schemas-microsoft-com:asm.v1}dependentAssembly',
    '{urn:schemas-microsoft-com:asm.v1}assemblyIdentity',
]

_MS_COMMON_CONTROLS_ATTRIBUTES = {
    "type": "win32",
    "name": "Microsoft.Windows.Common-Controls",
    "version": "6.0.0.0",
    "processorArchitecture": "*",
    "publicKeyToken": "6595b64144ccf1df",
    "language": "*",
}

#- Tests with default manifest template


# Generate default application manifest, which should be the same as the default XML template.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_default_manifest():
    import lxml
    import xmldiff.main

    app_manifest = winmanifest.create_application_manifest()

    tree_base = lxml.etree.fromstring(winmanifest._DEFAULT_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)

    # There should be no difference, as unmodified copy of template is returned.
    assert not diff


# Generate application manifest with uac_admin=True.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_default_manifest_uac_admin():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    app_manifest = winmanifest.create_application_manifest(uac_admin=True)

    tree_base = lxml.etree.fromstring(winmanifest._DEFAULT_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)

    # We expect exactly one change. The `level` attribute of `requestedExecutionLevel` element should be changed to
    # `requireAdministrator`.
    assert len(diff) == 1

    action = diff[0]
    assert isinstance(action, xmldiff.actions.UpdateAttrib)
    assert action.name == 'level'
    assert action.value == 'requireAdministrator'

    node = tree.xpath(action.node)[0]  # The node in xmldiff action is xpath
    assert _get_parent_tags(node) == _REQUEST_EXECUTION_LEVEL_TAGS


# Generate application manifest with uac_uiaccess=True.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_default_manifest_uac_uiaccess():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    app_manifest = winmanifest.create_application_manifest(uac_uiaccess=True)

    tree_base = lxml.etree.fromstring(winmanifest._DEFAULT_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)

    # We expect exactly one change. The `uiAccess` attribute of `requestedExecutionLevel` element should be changed to
    # `true`.
    assert len(diff) == 1

    action = diff[0]
    assert isinstance(action, xmldiff.actions.UpdateAttrib)
    assert action.name == 'uiAccess'
    assert action.value == 'true'

    node = tree.xpath(action.node)[0]  # The node in xmldiff action is xpath
    assert _get_parent_tags(node) == _REQUEST_EXECUTION_LEVEL_TAGS


# Generate application manifest with both uac_admin=True and uac_uiaccess=True.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_default_manifest_uac_admin_and_uiaccess():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    app_manifest = winmanifest.create_application_manifest(uac_admin=True, uac_uiaccess=True)

    tree_base = lxml.etree.fromstring(winmanifest._DEFAULT_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)

    # We expect exactly two changes. The `level` attribute of `requestedExecutionLevel` element should be changed to
    # `requireAdministrator` and the `uiAccess` attribute of `requestedExecutionLevel` element should be changed to
    # `true`.
    assert len(diff) == 2

    # First change action
    action = diff[0]
    assert isinstance(action, xmldiff.actions.UpdateAttrib)
    assert action.name == 'level'
    assert action.value == 'requireAdministrator'

    node = tree.xpath(action.node)[0]  # The node in xmldiff action is xpath
    assert _get_parent_tags(node) == _REQUEST_EXECUTION_LEVEL_TAGS

    # Second change action
    action = diff[1]
    assert isinstance(action, xmldiff.actions.UpdateAttrib)
    assert action.name == 'uiAccess'
    assert action.value == 'true'

    node = tree.xpath(action.node)[0]
    assert _get_parent_tags(node) == _REQUEST_EXECUTION_LEVEL_TAGS


#- Tests with custom manifest template


# Generate application manifest from template that has no <trustInfo> element, which we need to add to store the
# uac_admin and uac_uiaccess settings (even if left at default).
# The manifest also contains a `dpiAwareness` setting, which should be preserved (the old PyInstaller < 6.0 manifest
# handling code had no notion of this element, so it ended up striping it away).
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_custom_manifest_no_trust_info():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    _MANIFEST_XML = \
b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
    </application>
  </compatibility>
  <application xmlns="urn:schemas-microsoft-com:asm.v3">
    <windowsSettings>
      <longPathAware xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">true</longPathAware>
      <dpiAwareness xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">PerMonitorV2, unaware</dpiAwareness>
    </windowsSettings>
  </application>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity type="win32" name="Microsoft.Windows.Common-Controls" version="6.0.0.0" processorArchitecture="*" publicKeyToken="6595b64144ccf1df" language="*"></assemblyIdentity>
    </dependentAssembly>
  </dependency>
</assembly>
"""  # noqa: E128,E501

    app_manifest = winmanifest.create_application_manifest(_MANIFEST_XML)

    tree_base = lxml.etree.fromstring(_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)
    diff = _filter_whitespace_changes(diff)

    # After filtering newline/whitespace changes, modification still involves several steps.
    assert len(diff) == 6

    action = diff[0]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert tree.xpath(action.target)[0].tag == _REQUEST_EXECUTION_LEVEL_TAGS[0]
    assert action.tag == _REQUEST_EXECUTION_LEVEL_TAGS[1]

    action = diff[1]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _REQUEST_EXECUTION_LEVEL_TAGS[2]

    action = diff[2]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _REQUEST_EXECUTION_LEVEL_TAGS[3]

    action = diff[3]
    print(action)
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _REQUEST_EXECUTION_LEVEL_TAGS[4]

    action = diff[4]
    assert isinstance(action, xmldiff.actions.InsertAttrib)
    assert action.name == 'level'
    assert action.value == 'asInvoker'

    action = diff[5]
    assert isinstance(action, xmldiff.actions.InsertAttrib)
    assert action.name == 'uiAccess'
    assert action.value == 'false'


# Generate application manifest from template that has no dependent assembly for MS Common Controls, which we need to
# add.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_custom_manifest_no_ms_common_controls():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    _MANIFEST_XML = \
b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel level="asInvoker" uiAccess="false"></requestedExecutionLevel>
      </requestedPrivileges>
    </security>
  </trustInfo>
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
    </application>
  </compatibility>
  <application xmlns="urn:schemas-microsoft-com:asm.v3">
    <windowsSettings>
      <longPathAware xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">true</longPathAware>
      <dpiAwareness xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">PerMonitorV2, unaware</dpiAwareness>
    </windowsSettings>
  </application>
</assembly>
"""  # noqa: E128,E501

    app_manifest = winmanifest.create_application_manifest(_MANIFEST_XML)

    tree_base = lxml.etree.fromstring(_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)
    diff = _filter_whitespace_changes(diff)

    # After filtering newline/whitespace changes, modification still involves several steps.
    assert len(diff) == 9

    action = diff[0]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert tree.xpath(action.target)[0].tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[0]
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[1]

    action = diff[1]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[2]

    action = diff[2]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[3]

    # We do not care about insertion order, just that all attributes are added
    added_attributes = {}
    for action in diff[3:]:
        assert isinstance(action, xmldiff.actions.InsertAttrib)
        added_attributes[action.name] = action.value

    assert added_attributes == _MS_COMMON_CONTROLS_ATTRIBUTES


# Generate application manifest from template that has dependent assembly for MS Common Controls, but its attributes
# differ from our defaults.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_custom_manifest_different_ms_common_controls():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    _MANIFEST_XML = \
b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel level="asInvoker" uiAccess="false"></requestedExecutionLevel>
      </requestedPrivileges>
    </security>
  </trustInfo>
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
    </application>
  </compatibility>
  <application xmlns="urn:schemas-microsoft-com:asm.v3">
    <windowsSettings>
      <longPathAware xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">true</longPathAware>
      <dpiAwareness xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">PerMonitorV2, unaware</dpiAwareness>
    </windowsSettings>
  </application>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity type="win32" name="Microsoft.Windows.Common-Controls" version="5.0.0.0"></assemblyIdentity>
    </dependentAssembly>
  </dependency>
</assembly>
"""  # noqa: E128,E501

    app_manifest = winmanifest.create_application_manifest(_MANIFEST_XML)

    tree_base = lxml.etree.fromstring(_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)
    diff = _filter_whitespace_changes(diff)

    # After filtering newline/whitespace changes, modification still involves several steps:
    # one attribute modification (version) and three attribute additions.
    assert len(diff) == 4

    updated_attributes = {}
    added_attributes = {}
    for action in diff:
        if isinstance(action, xmldiff.actions.UpdateAttrib):
            updated_attributes[action.name] = action.value
        elif isinstance(action, xmldiff.actions.InsertAttrib):
            added_attributes[action.name] = action.value
        else:
            raise ValueError(f"Unexpected modification: {action}")

        assert tree.xpath(action.node)[0].tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[-1]

    assert updated_attributes == {'version': _MS_COMMON_CONTROLS_ATTRIBUTES['version']}
    assert added_attributes == {
        key: value
        for key, value in _MS_COMMON_CONTROLS_ATTRIBUTES.items()
        if key in ("processorArchitecture", "publicKeyToken", "language")
    }


# Generate application manifest from template that has dependent assembly for a custom library, but not for MS Common
# Controls. This is essentially the same as `test_manifest_custom_manifest_no_ms_common_controls`, but we should not
# touch the existing dependent assembly.
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_custom_manifest_no_ms_common_controls_with_custom_dependency():
    import lxml
    import xmldiff.main
    import xmldiff.actions

    _MANIFEST_XML = \
b"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel level="asInvoker" uiAccess="false"></requestedExecutionLevel>
      </requestedPrivileges>
    </security>
  </trustInfo>
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
    </application>
  </compatibility>
  <application xmlns="urn:schemas-microsoft-com:asm.v3">
    <windowsSettings>
      <longPathAware xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">true</longPathAware>
      <dpiAwareness xmlns="http://schemas.microsoft.com/SMI/2016/WindowsSettings">PerMonitorV2, unaware</dpiAwareness>
    </windowsSettings>
  </application>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity type="win32" name="MyAwesomeLibrary" version="1.0.0.0"></assemblyIdentity>
    </dependentAssembly>
  </dependency>
</assembly>
"""  # noqa: E128,E501

    app_manifest = winmanifest.create_application_manifest(_MANIFEST_XML)

    tree_base = lxml.etree.fromstring(_MANIFEST_XML)
    tree = lxml.etree.fromstring(app_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)
    diff = _filter_whitespace_changes(diff)

    # After filtering newline/whitespace changes, modification still involves several steps.
    assert len(diff) == 9

    action = diff[0]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert tree.xpath(action.target)[0].tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[0]
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[1]

    action = diff[1]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[2]

    action = diff[2]
    assert isinstance(action, xmldiff.actions.InsertNode)
    assert action.tag == _DEPENDENT_ASSEMBLY_IDENTITY_TAGS[3]

    # We do not care about insertion order, just that all attributes are added
    added_attributes = {}
    for action in diff[3:]:
        assert isinstance(action, xmldiff.actions.InsertAttrib)
        added_attributes[action.name] = action.value

    assert added_attributes == _MS_COMMON_CONTROLS_ATTRIBUTES


#- Test application manifest embedding and retrieval.
# Works only on Windows, because we need to test with actual executable with actual exeutable.


@pytest.mark.win32
def test_manifest_write_to_exe(tmp_path):
    # Locate bootloader executable
    bootloader_file = os.path.join(HOMEPATH, 'PyInstaller', 'bootloader', PLATFORM, 'run.exe')

    # Create a local copy
    test_file = str(tmp_path / 'test_file.exe')
    shutil.copyfile(bootloader_file, test_file)

    # Create manifest
    app_manifest = winmanifest.create_application_manifest(uac_admin=False, uac_uiaccess=True)

    # ... embed it, ...
    winmanifest.write_manifest_to_executable(test_file, app_manifest)

    # ... and read it back
    read_manifest = winmanifest.read_manifest_from_executable(test_file)

    # These should be identical, byte-for-byte
    assert read_manifest == app_manifest


# Similar to test_manifest_write_to_exe, but using custom manifest with non-ASCII characters.
@pytest.mark.win32
@importorskip('lxml')
@importorskip('xmldiff')
def test_manifest_write_to_exe_non_ascii_characters(tmp_path):
    import lxml
    import xmldiff.main
    import xmldiff.actions

    _MANIFEST_XML = \
"""<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
  <assemblyIdentity name="日本語で書かれた名前" processorArchitecture="amd64" type="win32" version="1.0.0.0"/>
  <dependency>
    <dependentAssembly>
      <assemblyIdentity language="*" name="Microsoft.Windows.Common-Controls" processorArchitecture="*" publicKeyToken="6595b64144ccf1df" type="win32" version="6.0.0.0"/>
    </dependentAssembly>
  </dependency>
  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
    <application>
      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"/>
      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"/>
      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"/>
      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"/>
      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"/>
    </application>
  </compatibility>
  <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
    <security>
      <requestedPrivileges>
        <requestedExecutionLevel level="asInvoker" uiAccess="false"/>
      </requestedPrivileges>
    </security>
  </trustInfo>
</assembly>
""".encode("utf-8")  # noqa: E128,E501

    # Locate bootloader executable
    bootloader_file = os.path.join(HOMEPATH, 'PyInstaller', 'bootloader', PLATFORM, 'run.exe')

    # Create a local copy
    test_file = str(tmp_path / 'test_file.exe')
    shutil.copyfile(bootloader_file, test_file)

    # Create manifest
    app_manifest = winmanifest.create_application_manifest(_MANIFEST_XML, uac_admin=False, uac_uiaccess=False)

    # ... embed it, ...
    winmanifest.write_manifest_to_executable(test_file, app_manifest)

    # ... and read it back
    read_manifest = winmanifest.read_manifest_from_executable(test_file)

    # These should be identical, byte-for-byte
    assert read_manifest == app_manifest

    # The retrieved manifest should have equivalent DOM to the original XML.
    # The raw string might differ from the original template due to closing tags.
    tree_base = lxml.etree.fromstring(_MANIFEST_XML)
    tree = lxml.etree.fromstring(read_manifest)

    diff = xmldiff.main.diff_trees(tree_base, tree)
    diff = _filter_whitespace_changes(diff)

    assert not diff



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import sys
from waflib.Tools import ar, d
from waflib.Configure import conf


@conf
def find_dmd(conf):
    conf.find_program(['dmd', 'dmd2', 'ldc'], var='D')
    out = conf.cmd_and_log(conf.env.D + ['--help'])
    if out.find("D Compiler v") == -1:
        out = conf.cmd_and_log(conf.env.D + ['-version'])
        if out.find("based on DMD v1.") == -1:
            conf.fatal("detected compiler is not dmd/ldc")


@conf
def common_flags_ldc(conf):
    v = conf.env
    v.DFLAGS = ['-d-version=Posix']
    v.LINKFLAGS = []
    v.DFLAGS_dshlib = ['-relocation-model=pic']


@conf
def common_flags_dmd(conf):
    v = conf.env
    v.D_SRC_F = ['-c']
    v.D_TGT_F = '-of%s'
    v.D_LINKER = v.D
    v.DLNK_SRC_F = ''
    v.DLNK_TGT_F = '-of%s'
    v.DINC_ST = '-I%s'
    v.DSHLIB_MARKER = v.DSTLIB_MARKER = ''
    v.DSTLIB_ST = v.DSHLIB_ST = '-L-l%s'
    v.DSTLIBPATH_ST = v.DLIBPATH_ST = '-L-L%s'
    v.LINKFLAGS_dprogram = ['-quiet']
    v.DFLAGS_dshlib = ['-fPIC']
    v.LINKFLAGS_dshlib = ['-L-shared']
    v.DHEADER_ext = '.di'
    v.DFLAGS_d_with_header = ['-H', '-Hf']
    v.D_HDR_F = '%s'


def configure(conf):
    conf.find_dmd()
    if sys.platform == 'win32':
        out = conf.cmd_and_log(conf.env.D + ['--help'])
        if out.find('D Compiler v2.') > -1:
            conf.fatal('dmd2 on Windows is not supported, use gdc or ldc2 instead')
    conf.load('ar')
    conf.load('d')
    conf.common_flags_dmd()
    conf.d_platform_flags()
    if str(conf.env.D).find('ldc') > -1:
        conf.common_flags_ldc()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, shutil
from waflib import Task, Utils, Errors, Node
from waflib.Configure import conf
from waflib.TaskGen import feature, before_method, after_method, taskgen_method
from waflib.Tools import ccroot

ccroot.USELIB_VARS['javac'] = set(['CLASSPATH', 'JAVACFLAGS'])
SOURCE_RE = '**/*.java'
JAR_RE = '**/*'
class_check_source = '''
public class Test {
	public static void main(String[] argv) {
		Class lib;
		if (argv.length < 1) {
			System.err.println("Missing argument");
			System.exit(77);
		}
		try {
			lib = Class.forName(argv[0]);
		} catch (ClassNotFoundException e) {
			System.err.println("ClassNotFoundException");
			System.exit(1);
		}
		lib = null;
		System.exit(0);
	}
}
'''


@feature('javac')
@before_method('process_source')
def apply_java(self):
    Utils.def_attrs(
        self, jarname='', classpath='', sourcepath='.', srcdir='.', jar_mf_attributes={}, jar_mf_classpath=[]
    )
    outdir = getattr(self, 'outdir', None)
    if outdir:
        if not isinstance(outdir, Node.Node):
            outdir = self.path.get_bld().make_node(self.outdir)
    else:
        outdir = self.path.get_bld()
    outdir.mkdir()
    self.outdir = outdir
    self.env.OUTDIR = outdir.abspath()
    self.javac_task = tsk = self.create_task('javac')
    tmp = []
    srcdir = getattr(self, 'srcdir', '')
    if isinstance(srcdir, Node.Node):
        srcdir = [srcdir]
    for x in Utils.to_list(srcdir):
        if isinstance(x, Node.Node):
            y = x
        else:
            y = self.path.find_dir(x)
            if not y:
                self.bld.fatal('Could not find the folder %s from %s' % (x, self.path))
        tmp.append(y)
    tsk.srcdir = tmp
    if getattr(self, 'compat', None):
        tsk.env.append_value('JAVACFLAGS', ['-source', str(self.compat)])
    if hasattr(self, 'sourcepath'):
        fold = [isinstance(x, Node.Node) and x or self.path.find_dir(x) for x in self.to_list(self.sourcepath)]
        names = os.pathsep.join([x.srcpath() for x in fold])
    else:
        names = [x.srcpath() for x in tsk.srcdir]
    if names:
        tsk.env.append_value('JAVACFLAGS', ['-sourcepath', names])


@taskgen_method
def java_use_rec(self, name, **kw):
    if name in self.tmp_use_seen:
        return
    self.tmp_use_seen.append(name)
    try:
        y = self.bld.get_tgen_by_name(name)
    except Errors.WafError:
        self.uselib.append(name)
        return
    else:
        y.post()
        if hasattr(y, 'jar_task'):
            self.use_lst.append(y.jar_task.outputs[0].abspath())
        else:
            if hasattr(y, 'outdir'):
                self.use_lst.append(y.outdir.abspath())
            else:
                self.use_lst.append(y.path.get_bld().abspath())
    for x in self.to_list(getattr(y, 'use', [])):
        self.java_use_rec(x)


@feature('javac')
@before_method('propagate_uselib_vars')
@after_method('apply_java')
def use_javac_files(self):
    self.use_lst = []
    self.tmp_use_seen = []
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    names = self.to_list(getattr(self, 'use', []))
    get = self.bld.get_tgen_by_name
    for x in names:
        try:
            tg = get(x)
        except Errors.WafError:
            self.uselib.append(x)
        else:
            tg.post()
            if hasattr(tg, 'jar_task'):
                self.use_lst.append(tg.jar_task.outputs[0].abspath())
                self.javac_task.set_run_after(tg.jar_task)
                self.javac_task.dep_nodes.extend(tg.jar_task.outputs)
            else:
                if hasattr(tg, 'outdir'):
                    base_node = tg.outdir
                else:
                    base_node = tg.path.get_bld()
                self.use_lst.append(base_node.abspath())
                self.javac_task.dep_nodes.extend([dx for dx in base_node.ant_glob(JAR_RE, remove=False, quiet=True)])
                for tsk in tg.tasks:
                    self.javac_task.set_run_after(tsk)
        if getattr(self, 'recurse_use', False) or self.bld.env.RECURSE_JAVA:
            self.java_use_rec(x)
    self.env.append_value('CLASSPATH', self.use_lst)


@feature('javac')
@after_method('apply_java', 'propagate_uselib_vars', 'use_javac_files')
def set_classpath(self):
    if getattr(self, 'classpath', None):
        self.env.append_unique('CLASSPATH', getattr(self, 'classpath', []))
    for x in self.tasks:
        x.env.CLASSPATH = os.pathsep.join(self.env.CLASSPATH) + os.pathsep


@feature('jar')
@after_method('apply_java', 'use_javac_files')
@before_method('process_source')
def jar_files(self):
    destfile = getattr(self, 'destfile', 'test.jar')
    jaropts = getattr(self, 'jaropts', [])
    manifest = getattr(self, 'manifest', None)
    basedir = getattr(self, 'basedir', None)
    if basedir:
        if not isinstance(self.basedir, Node.Node):
            basedir = self.path.get_bld().make_node(basedir)
    else:
        basedir = self.path.get_bld()
    if not basedir:
        self.bld.fatal('Could not find the basedir %r for %r' % (self.basedir, self))
    self.jar_task = tsk = self.create_task('jar_create')
    if manifest:
        jarcreate = getattr(self, 'jarcreate', 'cfm')
        if not isinstance(manifest, Node.Node):
            node = self.path.find_resource(manifest)
        else:
            node = manifest
        if not node:
            self.bld.fatal('invalid manifest file %r for %r' % (manifest, self))
        tsk.dep_nodes.append(node)
        jaropts.insert(0, node.abspath())
    else:
        jarcreate = getattr(self, 'jarcreate', 'cf')
    if not isinstance(destfile, Node.Node):
        destfile = self.path.find_or_declare(destfile)
    if not destfile:
        self.bld.fatal('invalid destfile %r for %r' % (destfile, self))
    tsk.set_outputs(destfile)
    tsk.basedir = basedir
    jaropts.append('-C')
    jaropts.append(basedir.bldpath())
    jaropts.append('.')
    tsk.env.JAROPTS = jaropts
    tsk.env.JARCREATE = jarcreate
    if getattr(self, 'javac_task', None):
        tsk.set_run_after(self.javac_task)


@feature('jar')
@after_method('jar_files')
def use_jar_files(self):
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    names = self.to_list(getattr(self, 'use', []))
    get = self.bld.get_tgen_by_name
    for x in names:
        try:
            y = get(x)
        except Errors.WafError:
            self.uselib.append(x)
        else:
            y.post()
            self.jar_task.run_after.update(y.tasks)


class JTask(Task.Task):
    def split_argfile(self, cmd):
        inline = [cmd[0]]
        infile = []
        for x in cmd[1:]:
            if x.startswith('-J'):
                inline.append(x)
            else:
                infile.append(self.quote_flag(x))
        return (inline, infile)


class jar_create(JTask):
    color = 'GREEN'
    run_str = '${JAR} ${JARCREATE} ${TGT} ${JAROPTS}'

    def runnable_status(self):
        for t in self.run_after:
            if not t.hasrun:
                return Task.ASK_LATER
        if not self.inputs:
            try:
                self.inputs = [
                    x for x in self.basedir.ant_glob(JAR_RE, remove=False, quiet=True) if id(x) != id(self.outputs[0])
                ]
            except Exception:
                raise Errors.WafError('Could not find the basedir %r for %r' % (self.basedir, self))
        return super(jar_create, self).runnable_status()


class javac(JTask):
    color = 'BLUE'
    run_str = '${JAVAC} -classpath ${CLASSPATH} -d ${OUTDIR} ${JAVACFLAGS} ${SRC}'
    vars = ['CLASSPATH', 'JAVACFLAGS', 'JAVAC', 'OUTDIR']

    def uid(self):
        lst = [self.__class__.__name__, self.generator.outdir.abspath()]
        for x in self.srcdir:
            lst.append(x.abspath())
        return Utils.h_list(lst)

    def runnable_status(self):
        for t in self.run_after:
            if not t.hasrun:
                return Task.ASK_LATER
        if not self.inputs:
            self.inputs = []
            for x in self.srcdir:
                if x.exists():
                    self.inputs.extend(x.ant_glob(SOURCE_RE, remove=False, quiet=True))
        return super(javac, self).runnable_status()

    def post_run(self):
        for node in self.generator.outdir.ant_glob('**/*.class', quiet=True):
            self.generator.bld.node_sigs[node] = self.uid()
        self.generator.bld.task_sigs[self.uid()] = self.cache_sig


@feature('javadoc')
@after_method('process_rule')
def create_javadoc(self):
    tsk = self.create_task('javadoc')
    tsk.classpath = getattr(self, 'classpath', [])
    self.javadoc_package = Utils.to_list(self.javadoc_package)
    if not isinstance(self.javadoc_output, Node.Node):
        self.javadoc_output = self.bld.path.find_or_declare(self.javadoc_output)


class javadoc(Task.Task):
    color = 'BLUE'

    def __str__(self):
        return '%s: %s -> %s\n' % (self.__class__.__name__, self.generator.srcdir, self.generator.javadoc_output)

    def run(self):
        env = self.env
        bld = self.generator.bld
        wd = bld.bldnode
        srcpath = self.generator.path.abspath() + os.sep + self.generator.srcdir
        srcpath += os.pathsep
        srcpath += self.generator.path.get_bld().abspath() + os.sep + self.generator.srcdir
        classpath = env.CLASSPATH
        classpath += os.pathsep
        classpath += os.pathsep.join(self.classpath)
        classpath = "".join(classpath)
        self.last_cmd = lst = []
        lst.extend(Utils.to_list(env.JAVADOC))
        lst.extend(['-d', self.generator.javadoc_output.abspath()])
        lst.extend(['-sourcepath', srcpath])
        lst.extend(['-classpath', classpath])
        lst.extend(['-subpackages'])
        lst.extend(self.generator.javadoc_package)
        lst = [x for x in lst if x]
        self.generator.bld.cmd_and_log(lst, cwd=wd, env=env.env or None, quiet=0)

    def post_run(self):
        nodes = self.generator.javadoc_output.ant_glob('**', quiet=True)
        for node in nodes:
            self.generator.bld.node_sigs[node] = self.uid()
        self.generator.bld.task_sigs[self.uid()] = self.cache_sig


def configure(self):
    java_path = self.environ['PATH'].split(os.pathsep)
    v = self.env
    if 'JAVA_HOME' in self.environ:
        java_path = [os.path.join(self.environ['JAVA_HOME'], 'bin')] + java_path
        self.env.JAVA_HOME = [self.environ['JAVA_HOME']]
    for x in 'javac java jar javadoc'.split():
        self.find_program(x, var=x.upper(), path_list=java_path, mandatory=(x not in ('javadoc')))
    if 'CLASSPATH' in self.environ:
        v.CLASSPATH = self.environ['CLASSPATH']
    if not v.JAR:
        self.fatal('jar is required for making java packages')
    if not v.JAVAC:
        self.fatal('javac is required for compiling java classes')
    v.JARCREATE = 'cf'
    v.JAVACFLAGS = []


@conf
def check_java_class(self, classname, with_classpath=None):
    javatestdir = '.waf-javatest'
    classpath = javatestdir
    if self.env.CLASSPATH:
        classpath += os.pathsep + self.env.CLASSPATH
    if isinstance(with_classpath, str):
        classpath += os.pathsep + with_classpath
    shutil.rmtree(javatestdir, True)
    os.mkdir(javatestdir)
    Utils.writef(os.path.join(javatestdir, 'Test.java'), class_check_source)
    self.exec_command(self.env.JAVAC + [os.path.join(javatestdir, 'Test.java')], shell=False)
    cmd = self.env.JAVA + ['-cp', classpath, 'Test', classname]
    self.to_log("%s\n" % str(cmd))
    found = self.exec_command(cmd, shell=False)
    self.msg('Checking for java class %s' % classname, not found)
    shutil.rmtree(javatestdir, True)
    return found


@conf
def check_jni_headers(conf):
    if not conf.env.CC_NAME and not conf.env.CXX_NAME:
        conf.fatal('load a compiler first (gcc, g++, ..)')
    if not conf.env.JAVA_HOME:
        conf.fatal('set JAVA_HOME in the system environment')
    javaHome = conf.env.JAVA_HOME[0]
    dir = conf.root.find_dir(conf.env.JAVA_HOME[0] + '/include')
    if dir is None:
        dir = conf.root.find_dir(conf.env.JAVA_HOME[0] + '/../Headers')
    if dir is None:
        conf.fatal('JAVA_HOME does not seem to be set properly')
    f = dir.ant_glob('**/(jni|jni_md).h')
    incDirs = [x.parent.abspath() for x in f]
    dir = conf.root.find_dir(conf.env.JAVA_HOME[0])
    f = dir.ant_glob('**/*jvm.(so|dll|dylib)')
    libDirs = [x.parent.abspath() for x in f] or [javaHome]
    f = dir.ant_glob('**/*jvm.(lib)')
    if f:
        libDirs = [[x, y.parent.abspath()] for x in libDirs for y in f]
    if conf.env.DEST_OS == 'freebsd':
        conf.env.append_unique('LINKFLAGS_JAVA', '-pthread')
    for d in libDirs:
        try:
            conf.check(
                header_name='jni.h',
                define_name='HAVE_JNI_H',
                lib='jvm',
                libpath=d,
                includes=incDirs,
                uselib_store='JAVA',
                uselib='JAVA'
            )
        except Exception:
            pass
        else:
            break
    else:
        conf.fatal('could not find lib jvm in %r (see config.log)' % libDirs)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Read and write resources from/to Win32 PE files.
"""

import PyInstaller.log as logging
from PyInstaller.compat import pywintypes, win32api

logger = logging.getLogger(__name__)

LOAD_LIBRARY_AS_DATAFILE = 2
ERROR_BAD_EXE_FORMAT = 193
ERROR_RESOURCE_DATA_NOT_FOUND = 1812
ERROR_RESOURCE_TYPE_NOT_FOUND = 1813
ERROR_RESOURCE_NAME_NOT_FOUND = 1814
ERROR_RESOURCE_LANG_NOT_FOUND = 1815


def get_resources(filename, types=None, names=None, languages=None):
    """
    Retrieve resources from the given PE file.

    filename: path to the PE file.
    types: a list of resource types (integers or strings) to search for (None = all).
    names: a list of resource names (integers or strings) to search for (None = all).
    languages: a list of resource languages (integers) to search for (None = all).

    Returns a dictionary of the form {type: {name: {language: data}}}, which might also be empty if no matching
    resources were found.
    """
    types = set(types) if types is not None else {"*"}
    names = set(names) if names is not None else {"*"}
    languages = set(languages) if languages is not None else {"*"}

    output = {}

    # Errors codes for which we swallow exceptions
    _IGNORE_EXCEPTIONS = {
        ERROR_RESOURCE_DATA_NOT_FOUND,
        ERROR_RESOURCE_TYPE_NOT_FOUND,
        ERROR_RESOURCE_NAME_NOT_FOUND,
        ERROR_RESOURCE_LANG_NOT_FOUND,
    }

    # Open file
    module_handle = win32api.LoadLibraryEx(filename, 0, LOAD_LIBRARY_AS_DATAFILE)

    # Enumerate available resource types
    try:
        available_types = win32api.EnumResourceTypes(module_handle)
    except pywintypes.error as e:
        if e.args[0] not in _IGNORE_EXCEPTIONS:
            raise
        available_types = []

    if "*" not in types:
        available_types = [res_type for res_type in available_types if res_type in types]

    for res_type in available_types:
        # Enumerate available names for the resource type.
        try:
            available_names = win32api.EnumResourceNames(module_handle, res_type)
        except pywintypes.error as e:
            if e.args[0] not in _IGNORE_EXCEPTIONS:
                raise
            continue

        if "*" not in names:
            available_names = [res_name for res_name in available_names if res_name in names]

        for res_name in available_names:
            # Enumerate available languages for the resource type and name combination.
            try:
                available_languages = win32api.EnumResourceLanguages(module_handle, res_type, res_name)
            except pywintypes.error as e:
                if e.args[0] not in _IGNORE_EXCEPTIONS:
                    raise
                continue

            if "*" not in languages:
                available_languages = [res_lang for res_lang in available_languages if res_lang in languages]

            for res_lang in available_languages:
                # Read data
                try:
                    data = win32api.LoadResource(module_handle, res_type, res_name, res_lang)
                except pywintypes.error as e:
                    if e.args[0] not in _IGNORE_EXCEPTIONS:
                        raise
                    continue

                if res_type not in output:
                    output[res_type] = {}
                if res_name not in output[res_type]:
                    output[res_type][res_name] = {}
                output[res_type][res_name][res_lang] = data

    # Close file
    win32api.FreeLibrary(module_handle)

    return output


def add_or_update_resource(filename, data, res_type, names=None, languages=None):
    """
    Update or add a single resource in the PE file with the given binary data.

    filename: path to the PE file.
    data: binary data to write to the resource.
    res_type: resource type to add/update (integer or string).
    names: a list of resource names (integers or strings) to update (None = all).
    languages: a list of resource languages (integers) to update (None = all).
    """
    if res_type == "*":
        raise ValueError("res_type cannot be a wildcard (*)!")

    names = set(names) if names is not None else {"*"}
    languages = set(languages) if languages is not None else {"*"}

    # Retrieve existing resources, filtered by the given resource type and given resource names and languages.
    resources = get_resources(filename, [res_type], names, languages)

    # Add res_type, name, language combinations that are not already present
    resources = resources.get(res_type, {})  # This is now a {name: {language: data}} dictionary

    for res_name in names:
        if res_name == "*":
            continue
        if res_name not in resources:
            resources[res_name] = {}

        for res_lang in languages:
            if res_lang == "*":
                continue
            if res_lang not in resources[res_name]:
                resources[res_name][res_lang] = None  # Just an indicator

    # Add resource to the target file, overwriting the existing resources with same type, name, language combinations.
    module_handle = win32api.BeginUpdateResource(filename, 0)
    for res_name in resources.keys():
        for res_lang in resources[res_name].keys():
            win32api.UpdateResource(module_handle, res_type, res_name, data, res_lang)
    win32api.EndUpdateResource(module_handle, 0)


def copy_resources_from_pe_file(filename, src_filename, types=None, names=None, languages=None):
    """
    Update or add resources in the given PE file by copying them over from the specified source PE file.

    filename: path to the PE file.
    src_filename: path to the source PE file.
    types: a list of resource types (integers or strings) to add/update via copy for (None = all).
    names: a list of resource names (integers or strings) to add/update via copy (None = all).
    languages: a list of resource languages (integers) to add/update via copy (None = all).
    """
    types = set(types) if types is not None else {"*"}
    names = set(names) if names is not None else {"*"}
    languages = set(languages) if languages is not None else {"*"}

    # Retrieve existing resources, filtered by the given resource type and given resource names and languages.
    resources = get_resources(src_filename, types, names, languages)

    for res_type, resources_for_type in resources.items():
        if "*" not in types and res_type not in types:
            continue
        for res_name, resources_for_type_name in resources_for_type.items():
            if "*" not in names and res_name not in names:
                continue
            for res_lang, data in resources_for_type_name.items():
                if "*" not in languages and res_lang not in languages:
                    continue
                add_or_update_resource(filename, data, res_type, [res_name], [res_lang])


def remove_all_resources(filename):
    """
    Remove all resources from the given PE file:
    """
    module_handle = win32api.BeginUpdateResource(filename, True)  # bDeleteExistingResources=True
    win32api.EndUpdateResource(module_handle, False)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from __future__ import with_statement
import os, re, shlex
from waflib import Build, Utils, Task, Options, Logs, Errors, Runner
from waflib.TaskGen import after_method, feature
from waflib.Configure import conf

WAF_CONFIG_H = 'config.h'
DEFKEYS = 'define_key'
INCKEYS = 'include_key'
SNIP_EMPTY_PROGRAM = '''
int main(int argc, char **argv) {
	(void)argc; (void)argv;
	return 0;
}
'''
MACRO_TO_DESTOS = {
    '__linux__': 'linux',
    '__GNU__': 'gnu',
    '__FreeBSD__': 'freebsd',
    '__NetBSD__': 'netbsd',
    '__OpenBSD__': 'openbsd',
    '__sun': 'sunos',
    '__hpux': 'hpux',
    '__sgi': 'irix',
    '_AIX': 'aix',
    '__CYGWIN__': 'cygwin',
    '__MSYS__': 'cygwin',
    '_UWIN': 'uwin',
    '_WIN64': 'win32',
    '_WIN32': 'win32',
    '__ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__': 'darwin',
    '__ENVIRONMENT_IPHONE_OS_VERSION_MIN_REQUIRED__': 'darwin',
    '__QNX__': 'qnx',
    '__native_client__': 'nacl'
}
MACRO_TO_DEST_CPU = {
    '__x86_64__': 'x86_64',
    '__amd64__': 'x86_64',
    '__i386__': 'x86',
    '__ia64__': 'ia',
    '__mips__': 'mips',
    '__sparc__': 'sparc',
    '__alpha__': 'alpha',
    '__aarch64__': 'aarch64',
    '__thumb__': 'thumb',
    '__arm__': 'arm',
    '__hppa__': 'hppa',
    '__powerpc__': 'powerpc',
    '__ppc__': 'powerpc',
    '__convex__': 'convex',
    '__m68k__': 'm68k',
    '__s390x__': 's390x',
    '__s390__': 's390',
    '__sh__': 'sh',
    '__xtensa__': 'xtensa',
    '__e2k__': 'e2k',
    '__riscv': 'riscv',
}


@conf
def parse_flags(self, line, uselib_store, env=None, force_static=False, posix=None):
    assert (isinstance(line, str))
    env = env or self.env
    if posix is None:
        posix = True
        if '\\' in line:
            posix = ('\\ ' in line) or ('\\\\' in line)
    lex = shlex.shlex(line, posix=posix)
    lex.whitespace_split = True
    lex.commenters = ''
    lst = list(lex)
    so_re = re.compile(r"\.so(?:\.[0-9]+)*$")
    uselib = uselib_store

    def app(var, val):
        env.append_value('%s_%s' % (var, uselib), val)

    def appu(var, val):
        env.append_unique('%s_%s' % (var, uselib), val)

    static = False
    while lst:
        x = lst.pop(0)
        st = x[:2]
        ot = x[2:]
        if st == '-I' or st == '/I':
            if not ot:
                ot = lst.pop(0)
            appu('INCLUDES', ot)
        elif st == '-i':
            tmp = [x, lst.pop(0)]
            app('CFLAGS', tmp)
            app('CXXFLAGS', tmp)
        elif st == '-D' or (env.CXX_NAME == 'msvc' and st == '/D'):
            if not ot:
                ot = lst.pop(0)
            app('DEFINES', ot)
        elif st == '-l':
            if not ot:
                ot = lst.pop(0)
            prefix = 'STLIB' if (force_static or static) else 'LIB'
            app(prefix, ot)
        elif st == '-L':
            if not ot:
                ot = lst.pop(0)
            prefix = 'STLIBPATH' if (force_static or static) else 'LIBPATH'
            appu(prefix, ot)
        elif x.startswith('/LIBPATH:'):
            prefix = 'STLIBPATH' if (force_static or static) else 'LIBPATH'
            appu(prefix, x.replace('/LIBPATH:', ''))
        elif x.startswith('-std='):
            prefix = 'CXXFLAGS' if '++' in x else 'CFLAGS'
            app(prefix, x)
        elif x.startswith('+') or x in ('-pthread', '-fPIC', '-fpic', '-fPIE', '-fpie', '-flto', '-fno-lto'):
            app('CFLAGS', x)
            app('CXXFLAGS', x)
            app('LINKFLAGS', x)
        elif x == '-framework':
            appu('FRAMEWORK', lst.pop(0))
        elif x.startswith('-F'):
            appu('FRAMEWORKPATH', x[2:])
        elif x == '-Wl,-rpath' or x == '-Wl,-R':
            app('RPATH', lst.pop(0).lstrip('-Wl,'))
        elif x.startswith('-Wl,-R,'):
            app('RPATH', x[7:])
        elif x.startswith('-Wl,-R'):
            app('RPATH', x[6:])
        elif x.startswith('-Wl,-rpath,'):
            app('RPATH', x[11:])
        elif x == '-Wl,-Bstatic' or x == '-Bstatic':
            static = True
        elif x == '-Wl,-Bdynamic' or x == '-Bdynamic':
            static = False
        elif x.startswith('-Wl') or x in ('-rdynamic', '-pie'):
            app('LINKFLAGS', x)
        elif x.startswith(('-m', '-f', '-dynamic', '-O', '-g')):
            app('CFLAGS', x)
            app('CXXFLAGS', x)
        elif x.startswith('-bundle'):
            app('LINKFLAGS', x)
        elif x.startswith(('-undefined', '-Xlinker')):
            arg = lst.pop(0)
            app('LINKFLAGS', [x, arg])
        elif x.startswith(('-arch', '-isysroot')):
            tmp = [x, lst.pop(0)]
            app('CFLAGS', tmp)
            app('CXXFLAGS', tmp)
            app('LINKFLAGS', tmp)
        elif x.endswith(('.a', '.dylib', '.lib')) or so_re.search(x):
            appu('LINKFLAGS', x)
        else:
            self.to_log('Unhandled flag %r' % x)


@conf
def validate_cfg(self, kw):
    if not 'path' in kw:
        if not self.env.PKGCONFIG:
            self.find_program('pkg-config', var='PKGCONFIG')
        kw['path'] = self.env.PKGCONFIG
    s = ('atleast_pkgconfig_version' in kw) + ('modversion' in kw) + ('package' in kw)
    if s != 1:
        raise ValueError('exactly one of atleast_pkgconfig_version, modversion and package must be set')
    if not 'msg' in kw:
        if 'atleast_pkgconfig_version' in kw:
            kw['msg'] = 'Checking for pkg-config version >= %r' % kw['atleast_pkgconfig_version']
        elif 'modversion' in kw:
            kw['msg'] = 'Checking for %r version' % kw['modversion']
        else:
            kw['msg'] = 'Checking for %r' % (kw['package'])
    if not 'okmsg' in kw and not 'modversion' in kw:
        kw['okmsg'] = 'yes'
    if not 'errmsg' in kw:
        kw['errmsg'] = 'not found'
    if 'atleast_pkgconfig_version' in kw:
        pass
    elif 'modversion' in kw:
        if not 'uselib_store' in kw:
            kw['uselib_store'] = kw['modversion']
        if not 'define_name' in kw:
            kw['define_name'] = '%s_VERSION' % Utils.quote_define_name(kw['uselib_store'])
    else:
        if not 'uselib_store' in kw:
            kw['uselib_store'] = Utils.to_list(kw['package'])[0].upper()
        if not 'define_name' in kw:
            kw['define_name'] = self.have_define(kw['uselib_store'])


@conf
def exec_cfg(self, kw):
    path = Utils.to_list(kw['path'])
    env = self.env.env or None
    if kw.get('pkg_config_path'):
        if not env:
            env = dict(self.environ)
        env['PKG_CONFIG_PATH'] = kw['pkg_config_path']

    def define_it():
        define_name = kw['define_name']
        if kw.get('global_define', 1):
            self.define(define_name, 1, False)
        else:
            self.env.append_unique('DEFINES_%s' % kw['uselib_store'], "%s=1" % define_name)
        if kw.get('add_have_to_env', 1):
            self.env[define_name] = 1

    if 'atleast_pkgconfig_version' in kw:
        cmd = path + ['--atleast-pkgconfig-version=%s' % kw['atleast_pkgconfig_version']]
        self.cmd_and_log(cmd, env=env)
        return
    if 'modversion' in kw:
        version = self.cmd_and_log(path + ['--modversion', kw['modversion']], env=env).strip()
        if not 'okmsg' in kw:
            kw['okmsg'] = version
        self.define(kw['define_name'], version)
        return version
    lst = [] + path
    defi = kw.get('define_variable')
    if not defi:
        defi = self.env.PKG_CONFIG_DEFINES or {}
    for key, val in defi.items():
        lst.append('--define-variable=%s=%s' % (key, val))
    static = kw.get('force_static', False)
    if 'args' in kw:
        args = Utils.to_list(kw['args'])
        if '--static' in args or '--static-libs' in args:
            static = True
        lst += args
    lst.extend(Utils.to_list(kw['package']))
    if 'variables' in kw:
        v_env = kw.get('env', self.env)
        vars = Utils.to_list(kw['variables'])
        for v in vars:
            val = self.cmd_and_log(lst + ['--variable=' + v], env=env).strip()
            var = '%s_%s' % (kw['uselib_store'], v)
            v_env[var] = val
        return
    ret = self.cmd_and_log(lst, env=env)
    define_it()
    self.parse_flags(ret, kw['uselib_store'], kw.get('env', self.env), force_static=static, posix=kw.get('posix'))
    return ret


@conf
def check_cfg(self, *k, **kw):
    self.validate_cfg(kw)
    if 'msg' in kw:
        self.start_msg(kw['msg'], **kw)
    ret = None
    try:
        ret = self.exec_cfg(kw)
    except self.errors.WafError as e:
        if 'errmsg' in kw:
            self.end_msg(kw['errmsg'], 'YELLOW', **kw)
        if Logs.verbose > 1:
            self.to_log('Command failure: %s' % e)
        self.fatal('The configuration failed')
    else:
        if not ret:
            ret = True
        kw['success'] = ret
        if 'okmsg' in kw:
            self.end_msg(self.ret_msg(kw['okmsg'], kw), **kw)
    return ret


def build_fun(bld):
    if bld.kw['compile_filename']:
        node = bld.srcnode.make_node(bld.kw['compile_filename'])
        node.write(bld.kw['code'])
    o = bld(features=bld.kw['features'], source=bld.kw['compile_filename'], target='testprog')
    for k, v in bld.kw.items():
        setattr(o, k, v)
    if not bld.kw.get('quiet'):
        bld.conf.to_log("==>\n%s\n<==" % bld.kw['code'])


@conf
def validate_c(self, kw):
    for x in ('type_name', 'field_name', 'function_name'):
        if x in kw:
            Logs.warn('Invalid argument %r in test' % x)
    if not 'build_fun' in kw:
        kw['build_fun'] = build_fun
    if not 'env' in kw:
        kw['env'] = self.env.derive()
    env = kw['env']
    if not 'compiler' in kw and not 'features' in kw:
        kw['compiler'] = 'c'
        if env.CXX_NAME and Task.classes.get('cxx'):
            kw['compiler'] = 'cxx'
            if not self.env.CXX:
                self.fatal('a c++ compiler is required')
        else:
            if not self.env.CC:
                self.fatal('a c compiler is required')
    if not 'compile_mode' in kw:
        kw['compile_mode'] = 'c'
        if 'cxx' in Utils.to_list(kw.get('features', [])) or kw.get('compiler') == 'cxx':
            kw['compile_mode'] = 'cxx'
    if not 'type' in kw:
        kw['type'] = 'cprogram'
    if not 'features' in kw:
        if not 'header_name' in kw or kw.get('link_header_test', True):
            kw['features'] = [kw['compile_mode'], kw['type']]
        else:
            kw['features'] = [kw['compile_mode']]
    else:
        kw['features'] = Utils.to_list(kw['features'])
    if not 'compile_filename' in kw:
        kw['compile_filename'] = 'test.c' + ((kw['compile_mode'] == 'cxx') and 'pp' or '')

    def to_header(dct):
        if 'header_name' in dct:
            dct = Utils.to_list(dct['header_name'])
            return ''.join(['#include <%s>\n' % x for x in dct])
        return ''

    if 'framework_name' in kw:
        fwkname = kw['framework_name']
        if not 'uselib_store' in kw:
            kw['uselib_store'] = fwkname.upper()
        if not kw.get('no_header'):
            fwk = '%s/%s.h' % (fwkname, fwkname)
            if kw.get('remove_dot_h'):
                fwk = fwk[:-2]
            val = kw.get('header_name', [])
            kw['header_name'] = Utils.to_list(val) + [fwk]
        kw['msg'] = 'Checking for framework %s' % fwkname
        kw['framework'] = fwkname
    elif 'header_name' in kw:
        if not 'msg' in kw:
            kw['msg'] = 'Checking for header %s' % kw['header_name']
        l = Utils.to_list(kw['header_name'])
        assert len(l), 'list of headers in header_name is empty'
        kw['code'] = to_header(kw) + SNIP_EMPTY_PROGRAM
        if not 'uselib_store' in kw:
            kw['uselib_store'] = l[0].upper()
        if not 'define_name' in kw:
            kw['define_name'] = self.have_define(l[0])
    if 'lib' in kw:
        if not 'msg' in kw:
            kw['msg'] = 'Checking for library %s' % kw['lib']
        if not 'uselib_store' in kw:
            kw['uselib_store'] = kw['lib'].upper()
    if 'stlib' in kw:
        if not 'msg' in kw:
            kw['msg'] = 'Checking for static library %s' % kw['stlib']
        if not 'uselib_store' in kw:
            kw['uselib_store'] = kw['stlib'].upper()
    if 'fragment' in kw:
        kw['code'] = kw['fragment']
        if not 'msg' in kw:
            kw['msg'] = 'Checking for code snippet'
        if not 'errmsg' in kw:
            kw['errmsg'] = 'no'
    for (flagsname, flagstype) in (('cxxflags', 'compiler'), ('cflags', 'compiler'), ('linkflags', 'linker')):
        if flagsname in kw:
            if not 'msg' in kw:
                kw['msg'] = 'Checking for %s flags %s' % (flagstype, kw[flagsname])
            if not 'errmsg' in kw:
                kw['errmsg'] = 'no'
    if not 'execute' in kw:
        kw['execute'] = False
    if kw['execute']:
        kw['features'].append('test_exec')
        kw['chmod'] = Utils.O755
    if not 'errmsg' in kw:
        kw['errmsg'] = 'not found'
    if not 'okmsg' in kw:
        kw['okmsg'] = 'yes'
    if not 'code' in kw:
        kw['code'] = SNIP_EMPTY_PROGRAM
    if self.env[INCKEYS]:
        kw['code'] = '\n'.join(['#include <%s>' % x for x in self.env[INCKEYS]]) + '\n' + kw['code']
    if kw.get('merge_config_header') or env.merge_config_header:
        kw['code'] = '%s\n\n%s' % (self.get_config_header(), kw['code'])
        env.DEFINES = []
    if not kw.get('success'):
        kw['success'] = None
    if 'define_name' in kw:
        self.undefine(kw['define_name'])
    if not 'msg' in kw:
        self.fatal('missing "msg" in conf.check(...)')


@conf
def post_check(self, *k, **kw):
    is_success = 0
    if kw['execute']:
        if kw['success'] is not None:
            if kw.get('define_ret'):
                is_success = kw['success']
            else:
                is_success = (kw['success'] == 0)
    else:
        is_success = (kw['success'] == 0)
    if kw.get('define_name'):
        comment = kw.get('comment', '')
        define_name = kw['define_name']
        if kw['execute'] and kw.get('define_ret') and isinstance(is_success, str):
            if kw.get('global_define', 1):
                self.define(define_name, is_success, quote=kw.get('quote', 1), comment=comment)
            else:
                if kw.get('quote', 1):
                    succ = '"%s"' % is_success
                else:
                    succ = int(is_success)
                val = '%s=%s' % (define_name, succ)
                var = 'DEFINES_%s' % kw['uselib_store']
                self.env.append_value(var, val)
        else:
            if kw.get('global_define', 1):
                self.define_cond(define_name, is_success, comment=comment)
            else:
                var = 'DEFINES_%s' % kw['uselib_store']
                self.env.append_value(var, '%s=%s' % (define_name, int(is_success)))
        if kw.get('add_have_to_env', 1):
            if kw.get('uselib_store'):
                self.env[self.have_define(kw['uselib_store'])] = 1
            elif kw['execute'] and kw.get('define_ret'):
                self.env[define_name] = is_success
            else:
                self.env[define_name] = int(is_success)
    if 'header_name' in kw:
        if kw.get('auto_add_header_name'):
            self.env.append_value(INCKEYS, Utils.to_list(kw['header_name']))
    if is_success and 'uselib_store' in kw:
        from waflib.Tools import ccroot
        _vars = set()
        for x in kw['features']:
            if x in ccroot.USELIB_VARS:
                _vars |= ccroot.USELIB_VARS[x]
        for k in _vars:
            x = k.lower()
            if x in kw:
                self.env.append_value(k + '_' + kw['uselib_store'], kw[x])
    return is_success


@conf
def check(self, *k, **kw):
    self.validate_c(kw)
    self.start_msg(kw['msg'], **kw)
    ret = None
    try:
        ret = self.run_build(*k, **kw)
    except self.errors.ConfigurationError:
        self.end_msg(kw['errmsg'], 'YELLOW', **kw)
        if Logs.verbose > 1:
            raise
        else:
            self.fatal('The configuration failed')
    else:
        kw['success'] = ret
    ret = self.post_check(*k, **kw)
    if not ret:
        self.end_msg(kw['errmsg'], 'YELLOW', **kw)
        self.fatal('The configuration failed %r' % ret)
    else:
        self.end_msg(self.ret_msg(kw['okmsg'], kw), **kw)
    return ret


class test_exec(Task.Task):
    color = 'PINK'

    def run(self):
        cmd = [self.inputs[0].abspath()] + getattr(self.generator, 'test_args', [])
        if getattr(self.generator, 'rpath', None):
            if getattr(self.generator, 'define_ret', False):
                self.generator.bld.retval = self.generator.bld.cmd_and_log(cmd)
            else:
                self.generator.bld.retval = self.generator.bld.exec_command(cmd)
        else:
            env = self.env.env or {}
            env.update(dict(os.environ))
            for var in ('LD_LIBRARY_PATH', 'DYLD_LIBRARY_PATH', 'PATH'):
                env[var] = self.inputs[0].parent.abspath() + os.path.pathsep + env.get(var, '')
            if getattr(self.generator, 'define_ret', False):
                self.generator.bld.retval = self.generator.bld.cmd_and_log(cmd, env=env)
            else:
                self.generator.bld.retval = self.generator.bld.exec_command(cmd, env=env)


@feature('test_exec')
@after_method('apply_link')
def test_exec_fun(self):
    self.create_task('test_exec', self.link_task.outputs[0])


@conf
def check_cxx(self, *k, **kw):
    kw['compiler'] = 'cxx'
    return self.check(*k, **kw)


@conf
def check_cc(self, *k, **kw):
    kw['compiler'] = 'c'
    return self.check(*k, **kw)


@conf
def set_define_comment(self, key, comment):
    coms = self.env.DEFINE_COMMENTS
    if not coms:
        coms = self.env.DEFINE_COMMENTS = {}
    coms[key] = comment or ''


@conf
def get_define_comment(self, key):
    coms = self.env.DEFINE_COMMENTS or {}
    return coms.get(key, '')


@conf
def define(self, key, val, quote=True, comment=''):
    assert isinstance(key, str)
    if not key:
        return
    if val is True:
        val = 1
    elif val in (False, None):
        val = 0
    if isinstance(val, int) or isinstance(val, float):
        s = '%s=%s'
    else:
        s = quote and '%s="%s"' or '%s=%s'
    app = s % (key, str(val))
    ban = key + '='
    lst = self.env.DEFINES
    for x in lst:
        if x.startswith(ban):
            lst[lst.index(x)] = app
            break
    else:
        self.env.append_value('DEFINES', app)
    self.env.append_unique(DEFKEYS, key)
    self.set_define_comment(key, comment)


@conf
def undefine(self, key, comment=''):
    assert isinstance(key, str)
    if not key:
        return
    ban = key + '='
    lst = [x for x in self.env.DEFINES if not x.startswith(ban)]
    self.env.DEFINES = lst
    self.env.append_unique(DEFKEYS, key)
    self.set_define_comment(key, comment)


@conf
def define_cond(self, key, val, comment=''):
    assert isinstance(key, str)
    if not key:
        return
    if val:
        self.define(key, 1, comment=comment)
    else:
        self.undefine(key, comment=comment)


@conf
def is_defined(self, key):
    assert key and isinstance(key, str)
    ban = key + '='
    for x in self.env.DEFINES:
        if x.startswith(ban):
            return True
    return False


@conf
def get_define(self, key):
    assert key and isinstance(key, str)
    ban = key + '='
    for x in self.env.DEFINES:
        if x.startswith(ban):
            return x[len(ban):]
    return None


@conf
def have_define(self, key):
    return (self.env.HAVE_PAT or 'HAVE_%s') % Utils.quote_define_name(key)


@conf
def write_config_header(
    self, configfile='', guard='', top=False, defines=True, headers=False, remove=True, define_prefix=''
):
    if not configfile:
        configfile = WAF_CONFIG_H
    waf_guard = guard or 'W_%s_WAF' % Utils.quote_define_name(configfile)
    node = top and self.bldnode or self.path.get_bld()
    node = node.make_node(configfile)
    node.parent.mkdir()
    lst = ['/* WARNING! All changes made to this file will be lost! */\n']
    lst.append('#ifndef %s\n#define %s\n' % (waf_guard, waf_guard))
    lst.append(self.get_config_header(defines, headers, define_prefix=define_prefix))
    lst.append('\n#endif /* %s */\n' % waf_guard)
    node.write('\n'.join(lst))
    self.env.append_unique(Build.CFG_FILES, [node.abspath()])
    if remove:
        for key in self.env[DEFKEYS]:
            self.undefine(key)
        self.env[DEFKEYS] = []


@conf
def get_config_header(self, defines=True, headers=False, define_prefix=''):
    lst = []
    if self.env.WAF_CONFIG_H_PRELUDE:
        lst.append(self.env.WAF_CONFIG_H_PRELUDE)
    if headers:
        for x in self.env[INCKEYS]:
            lst.append('#include <%s>' % x)
    if defines:
        tbl = {}
        for k in self.env.DEFINES:
            a, _, b = k.partition('=')
            tbl[a] = b
        for k in self.env[DEFKEYS]:
            caption = self.get_define_comment(k)
            if caption:
                caption = ' /* %s */' % caption
            try:
                txt = '#define %s%s %s%s' % (define_prefix, k, tbl[k], caption)
            except KeyError:
                txt = '/* #undef %s%s */%s' % (define_prefix, k, caption)
            lst.append(txt)
    return "\n".join(lst)


@conf
def cc_add_flags(conf):
    conf.add_os_flags('CPPFLAGS', dup=False)
    conf.add_os_flags('CFLAGS', dup=False)


@conf
def cxx_add_flags(conf):
    conf.add_os_flags('CPPFLAGS', dup=False)
    conf.add_os_flags('CXXFLAGS', dup=False)


@conf
def link_add_flags(conf):
    conf.add_os_flags('LINKFLAGS', dup=False)
    conf.add_os_flags('LDFLAGS', dup=False)


@conf
def cc_load_tools(conf):
    if not conf.env.DEST_OS:
        conf.env.DEST_OS = Utils.unversioned_sys_platform()
    conf.load('c')


@conf
def cxx_load_tools(conf):
    if not conf.env.DEST_OS:
        conf.env.DEST_OS = Utils.unversioned_sys_platform()
    conf.load('cxx')


@conf
def get_cc_version(conf, cc, gcc=False, icc=False, clang=False):
    cmd = cc + ['-dM', '-E', '-']
    env = conf.env.env or None
    try:
        out, err = conf.cmd_and_log(cmd, output=0, input='\n'.encode(), env=env)
    except Errors.WafError:
        conf.fatal('Could not determine the compiler version %r' % cmd)
    if gcc:
        if out.find('__INTEL_COMPILER') >= 0:
            conf.fatal('The intel compiler pretends to be gcc')
        if out.find('__GNUC__') < 0 and out.find('__clang__') < 0:
            conf.fatal('Could not determine the compiler type')
    if icc and out.find('__INTEL_COMPILER') < 0:
        conf.fatal('Not icc/icpc')
    if clang and out.find('__clang__') < 0:
        conf.fatal('Not clang/clang++')
    if not clang and out.find('__clang__') >= 0:
        conf.fatal('Could not find gcc/g++ (only Clang), if renamed try eg: CC=gcc48 CXX=g++48 waf configure')
    k = {}
    if icc or gcc or clang:
        out = out.splitlines()
        for line in out:
            lst = shlex.split(line)
            if len(lst) > 2:
                key = lst[1]
                val = lst[2]
                k[key] = val

        def isD(var):
            return var in k

        if not conf.env.DEST_OS:
            conf.env.DEST_OS = ''
        for i in MACRO_TO_DESTOS:
            if isD(i):
                conf.env.DEST_OS = MACRO_TO_DESTOS[i]
                break
        else:
            if isD('__APPLE__') and isD('__MACH__'):
                conf.env.DEST_OS = 'darwin'
            elif isD('__unix__'):
                conf.env.DEST_OS = 'generic'
        if isD('__ELF__'):
            conf.env.DEST_BINFMT = 'elf'
        elif isD('__WINNT__') or isD('__CYGWIN__') or isD('_WIN32'):
            conf.env.DEST_BINFMT = 'pe'
            if not conf.env.IMPLIBDIR:
                conf.env.IMPLIBDIR = conf.env.LIBDIR
            conf.env.LIBDIR = conf.env.BINDIR
        elif isD('__APPLE__'):
            conf.env.DEST_BINFMT = 'mac-o'
        if not conf.env.DEST_BINFMT:
            conf.env.DEST_BINFMT = Utils.destos_to_binfmt(conf.env.DEST_OS)
        for i in MACRO_TO_DEST_CPU:
            if isD(i):
                conf.env.DEST_CPU = MACRO_TO_DEST_CPU[i]
                break
        Logs.debug(
            'ccroot: dest platform: ' + ' '.join([conf.env[x] or '?' for x in ('DEST_OS', 'DEST_BINFMT', 'DEST_CPU')])
        )
        if icc:
            ver = k['__INTEL_COMPILER']
            conf.env.CC_VERSION = (ver[:-2], ver[-2], ver[-1])
        else:
            if isD('__clang__') and isD('__clang_major__'):
                conf.env.CC_VERSION = (k['__clang_major__'], k['__clang_minor__'], k['__clang_patchlevel__'])
            else:
                conf.env.CC_VERSION = (k['__GNUC__'], k['__GNUC_MINOR__'], k.get('__GNUC_PATCHLEVEL__', '0'))
    return k


@conf
def get_xlc_version(conf, cc):
    cmd = cc + ['-qversion']
    try:
        out, err = conf.cmd_and_log(cmd, output=0)
    except Errors.WafError:
        conf.fatal('Could not find xlc %r' % cmd)
    for v in (r"IBM XL C/C\+\+.* V(?P<major>\d*)\.(?P<minor>\d*)",):
        version_re = re.compile(v, re.I).search
        match = version_re(out or err)
        if match:
            k = match.groupdict()
            conf.env.CC_VERSION = (k['major'], k['minor'])
            break
    else:
        conf.fatal('Could not determine the XLC version.')


@conf
def get_suncc_version(conf, cc):
    cmd = cc + ['-V']
    try:
        out, err = conf.cmd_and_log(cmd, output=0)
    except Errors.WafError as e:
        if not (hasattr(e, 'returncode') and hasattr(e, 'stdout') and hasattr(e, 'stderr')):
            conf.fatal('Could not find suncc %r' % cmd)
        out = e.stdout
        err = e.stderr
    version = (out or err)
    version = version.splitlines()[0]
    version_re = re.compile(
        r'cc: (studio.*?|\s+)?(sun\s+(c\+\+|c)|(WorkShop\s+Compilers))?\s+(?P<major>\d*)\.(?P<minor>\d*)', re.I
    ).search
    match = version_re(version)
    if match:
        k = match.groupdict()
        conf.env.CC_VERSION = (k['major'], k['minor'])
    else:
        conf.fatal('Could not determine the suncc version.')


@conf
def add_as_needed(self):
    if self.env.DEST_BINFMT == 'elf' and 'gcc' in (self.env.CXX_NAME, self.env.CC_NAME):
        self.env.append_unique('LINKFLAGS', '-Wl,--as-needed')


class cfgtask(Task.Task):
    def __init__(self, *k, **kw):
        Task.Task.__init__(self, *k, **kw)
        self.run_after = set()

    def display(self):
        return ''

    def runnable_status(self):
        for x in self.run_after:
            if not x.hasrun:
                return Task.ASK_LATER
        return Task.RUN_ME

    def uid(self):
        return Utils.SIG_NIL

    def signature(self):
        return Utils.SIG_NIL

    def run(self):
        conf = self.conf
        bld = Build.BuildContext(top_dir=conf.srcnode.abspath(), out_dir=conf.bldnode.abspath())
        bld.env = conf.env
        bld.init_dirs()
        bld.in_msg = 1
        bld.logger = self.logger
        bld.multicheck_task = self
        args = self.args
        try:
            if 'func' in args:
                bld.test(
                    build_fun=args['func'],
                    msg=args.get('msg', ''),
                    okmsg=args.get('okmsg', ''),
                    errmsg=args.get('errmsg', ''),
                )
            else:
                args['multicheck_mandatory'] = args.get('mandatory', True)
                args['mandatory'] = True
                try:
                    bld.check(**args)
                finally:
                    args['mandatory'] = args['multicheck_mandatory']
        except Exception:
            return 1

    def process(self):
        Task.Task.process(self)
        if 'msg' in self.args:
            with self.generator.bld.multicheck_lock:
                self.conf.start_msg(self.args['msg'])
                if self.hasrun == Task.NOT_RUN:
                    self.conf.end_msg('test cancelled', 'YELLOW')
                elif self.hasrun != Task.SUCCESS:
                    self.conf.end_msg(self.args.get('errmsg', 'no'), 'YELLOW')
                else:
                    self.conf.end_msg(self.args.get('okmsg', 'yes'), 'GREEN')


@conf
def multicheck(self, *k, **kw):
    self.start_msg(kw.get('msg', 'Executing %d configuration tests' % len(k)), **kw)
    for var in ('DEFINES', DEFKEYS):
        self.env.append_value(var, [])
    self.env.DEFINE_COMMENTS = self.env.DEFINE_COMMENTS or {}

    class par(object):
        def __init__(self):
            self.keep = False
            self.task_sigs = {}
            self.progress_bar = 0

        def total(self):
            return len(tasks)

        def to_log(self, *k, **kw):
            return

    bld = par()
    bld.keep = kw.get('run_all_tests', True)
    bld.imp_sigs = {}
    tasks = []
    id_to_task = {}
    for counter, dct in enumerate(k):
        x = Task.classes['cfgtask'](bld=bld, env=None)
        tasks.append(x)
        x.args = dct
        x.args['multicheck_counter'] = counter
        x.bld = bld
        x.conf = self
        x.args = dct
        x.logger = Logs.make_mem_logger(str(id(x)), self.logger)
        if 'id' in dct:
            id_to_task[dct['id']] = x
    for x in tasks:
        for key in Utils.to_list(x.args.get('before_tests', [])):
            tsk = id_to_task[key]
            if not tsk:
                raise ValueError('No test named %r' % key)
            tsk.run_after.add(x)
        for key in Utils.to_list(x.args.get('after_tests', [])):
            tsk = id_to_task[key]
            if not tsk:
                raise ValueError('No test named %r' % key)
            x.run_after.add(tsk)

    def it():
        yield tasks
        while 1:
            yield []

    bld.producer = p = Runner.Parallel(bld, Options.options.jobs)
    bld.multicheck_lock = Utils.threading.Lock()
    p.biter = it()
    self.end_msg('started')
    p.start()
    for x in tasks:
        x.logger.memhandler.flush()
    self.start_msg('-> processing test results')
    if p.error:
        for x in p.error:
            if getattr(x, 'err_msg', None):
                self.to_log(x.err_msg)
                self.end_msg('fail', color='RED')
                raise Errors.WafError('There is an error in the library, read config.log for more information')
    failure_count = 0
    for x in tasks:
        if x.hasrun not in (Task.SUCCESS, Task.NOT_RUN):
            failure_count += 1
    if failure_count:
        self.end_msg(kw.get('errmsg', '%s test failed' % failure_count), color='YELLOW', **kw)
    else:
        self.end_msg('all ok', **kw)
    for x in tasks:
        if x.hasrun != Task.SUCCESS:
            if x.args.get('mandatory', True):
                self.fatal(kw.get('fatalmsg') or 'One of the tests has failed, read config.log for more information')


@conf
def check_gcc_o_space(self, mode='c'):
    if int(self.env.CC_VERSION[0]) > 4:
        return
    self.env.stash()
    if mode == 'c':
        self.env.CCLNK_TGT_F = ['-o', '']
    elif mode == 'cxx':
        self.env.CXXLNK_TGT_F = ['-o', '']
    features = '%s %sshlib' % (mode, mode)
    try:
        self.check(
            msg='Checking if the -o link must be split from arguments', fragment=SNIP_EMPTY_PROGRAM, features=features
        )
    except self.errors.ConfigurationError:
        self.env.revert()
    else:
        self.env.commit()



"""
Test for import machinery
"""
import unittest
import sys
import textwrap
import subprocess
import os
from PyInstaller.lib.modulegraph import modulegraph


class TestModuleGraphImport (unittest.TestCase):
    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        def assertIsInstance(self, value, types):
            if not isinstance(value, types):
                self.fail("%r is not an instance of %r"%(value, types))

    def test_compat(self):
        root = os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                'testpkg-compatmodule')
        mf = modulegraph.ModuleGraph(path=[ root ] + sys.path)
        mf.import_hook('pkg.api')

        node = mf.find_node('pkg')
        self.assertIsInstance(node, modulegraph.Package)

        node = mf.find_node('pkg.api')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mf.find_node('pkg.api2')
        self.assertIsInstance(node, modulegraph.InvalidSourceModule)

        node = mf.find_node('pkg.api3')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mf.find_node('http.client')
        self.assertIsInstance(node, modulegraph.SourceModule)

        node = mf.find_node('urllib2')
        self.assertIs(node, None)


if __name__ == "__main__":
    unittest.main()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Encode a string into utf8 bytestring and decode it back. The resulting string should match the original.

import codecs

str_a = 'foo bar fóó bář, fěě, ďěž'
str_a_utf8 = codecs.getencoder('utf-8')(str_a)[0]
str_b = codecs.getdecoder('utf-8')(str_a_utf8)[0]

print('codecs working: %s' % (str_a == str_b))
assert str_a == str_b



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Task
from waflib.TaskGen import extension
from waflib.Tools import c_preproc


@extension('.rc')
def rc_file(self, node):
    obj_ext = '.rc.o'
    if self.env.WINRC_TGT_F == '/fo':
        obj_ext = '.res'
    rctask = self.create_task('winrc', node, node.change_ext(obj_ext))
    try:
        self.compiled_tasks.append(rctask)
    except AttributeError:
        self.compiled_tasks = [rctask]


re_lines = re.compile(
    r'(?:^[ \t]*(#|%:)[ \t]*(ifdef|ifndef|if|else|elif|endif|include|import|define|undef|pragma)[ \t]*(.*?)\s*$)|'
    r'(?:^\w+[ \t]*(ICON|BITMAP|CURSOR|HTML|FONT|MESSAGETABLE|TYPELIB|REGISTRY|D3DFX)[ \t]*(.*?)\s*$)',
    re.IGNORECASE | re.MULTILINE
)


class rc_parser(c_preproc.c_parser):
    def filter_comments(self, node):
        code = node.read()
        if c_preproc.use_trigraphs:
            for (a, b) in c_preproc.trig_def:
                code = code.split(a).join(b)
        code = c_preproc.re_nl.sub('', code)
        code = c_preproc.re_cpp.sub(c_preproc.repl, code)
        ret = []
        for m in re.finditer(re_lines, code):
            if m.group(2):
                ret.append((m.group(2), m.group(3)))
            else:
                ret.append(('include', m.group(5)))
        return ret


class winrc(Task.Task):
    run_str = '${WINRC} ${WINRCFLAGS} ${CPPPATH_ST:INCPATHS} ${DEFINES_ST:DEFINES} ${WINRC_TGT_F} ${TGT} ${WINRC_SRC_F} ${SRC}'
    color = 'BLUE'

    def scan(self):
        tmp = rc_parser(self.generator.includes_nodes)
        tmp.start(self.inputs[0], self.env)
        return (tmp.nodes, tmp.names)


def configure(conf):
    v = conf.env
    if not v.WINRC:
        if v.CC_NAME == 'msvc':
            conf.find_program('RC', var='WINRC', path_list=v.PATH)
            v.WINRC_TGT_F = '/fo'
            v.WINRC_SRC_F = ''
        else:
            conf.find_program('windres', var='WINRC', path_list=v.PATH)
            v.WINRC_TGT_F = '-o'
            v.WINRC_SRC_F = '-i'



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

sys.path.insert(0, os.path.dirname(__file__))

# Hack that prevents PyInstaller.compat from failing due to unmet run-time dependencies (importlib-metadata on
# python < 3.10, pywin32-ctypes on Windows). These dependencies are not required for the subset of functionality that is
# used here.
os.environ["_PYINSTALLER_SETUP"] = "1"


class CustomBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        # Inject the platform specifier into the wheel's filename.
        if os.environ.get("PYI_WHEEL_TAG"):
            build_data["tag"] = "py3-none-" + os.environ["PYI_WHEEL_TAG"]

        pyi_platform = os.environ.get("PYI_PLATFORM")
        if pyi_platform:
            if "Darwin" in pyi_platform:
                icons = ["icns"]
            elif "Windows" in pyi_platform:
                icons = ["ico"]
            else:
                icons = []
        else:
            icons = ["ico", "icns"]

        build_data["artifacts"] += [
            f"PyInstaller/bootloader/{pyi_platform or '*'}/*",
            *(f"PyInstaller/bootloader/images/*.{suffix}" for suffix in icons),
        ]
        self.run()

    def bootloader_exists(self):
        # Checks if the console, non-debug bootloader exists
        from PyInstaller import HOMEPATH, PLATFORM
        exe = 'run'
        pyi_platform = os.environ.get("PYI_PLATFORM", PLATFORM)
        if "Windows" in pyi_platform:
            exe = 'run.exe'
        exe = os.path.join(HOMEPATH, 'PyInstaller', 'bootloader', pyi_platform, exe)
        return os.path.isfile(exe)

    def compile_bootloader(self):
        import subprocess
        from PyInstaller import HOMEPATH

        src_dir = os.path.join(HOMEPATH, 'bootloader')
        additional_args = os.getenv('PYINSTALLER_BOOTLOADER_WAF_ARGS', '').strip().split()
        cmd = [sys.executable, './waf', 'configure', 'all']
        cmd += additional_args
        rc = subprocess.call(cmd, cwd=src_dir)
        if rc:
            raise SystemExit('ERROR: Failed compiling the bootloader. Please compile manually and rerun')

    def run(self):
        if self.bootloader_exists() and not os.environ.get("PYINSTALLER_COMPILE_BOOTLOADER"):
            return
        print(
            'No precompiled bootloader found or compile forced. Trying to compile the bootloader for you ...',
            file=sys.stderr
        )
        self.compile_bootloader()
        if not self.bootloader_exists():
            raise SystemExit("ERROR: Bootloaders have been compiled for the wrong platform")



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re
from waflib import Utils
from waflib.Tools import fc, fc_config, fc_scan, ar
from waflib.Configure import conf


@conf
def find_gfortran(conf):
    fc = conf.find_program(['gfortran', 'g77'], var='FC')
    conf.get_gfortran_version(fc)
    conf.env.FC_NAME = 'GFORTRAN'


@conf
def gfortran_flags(conf):
    v = conf.env
    v.FCFLAGS_fcshlib = ['-fPIC']
    v.FORTRANMODFLAG = ['-J', '']
    v.FCFLAGS_DEBUG = ['-Werror']


@conf
def gfortran_modifier_win32(conf):
    fc_config.fortran_modifier_win32(conf)


@conf
def gfortran_modifier_cygwin(conf):
    fc_config.fortran_modifier_cygwin(conf)


@conf
def gfortran_modifier_darwin(conf):
    fc_config.fortran_modifier_darwin(conf)


@conf
def gfortran_modifier_platform(conf):
    dest_os = conf.env.DEST_OS or Utils.unversioned_sys_platform()
    gfortran_modifier_func = getattr(conf, 'gfortran_modifier_' + dest_os, None)
    if gfortran_modifier_func:
        gfortran_modifier_func()


@conf
def get_gfortran_version(conf, fc):
    version_re = re.compile(r"GNU\s*Fortran", re.I).search
    cmd = fc + ['--version']
    out, err = fc_config.getoutput(conf, cmd, stdin=False)
    if out:
        match = version_re(out)
    else:
        match = version_re(err)
    if not match:
        conf.fatal('Could not determine the compiler type')
    cmd = fc + ['-dM', '-E', '-']
    out, err = fc_config.getoutput(conf, cmd, stdin=True)
    if out.find('__GNUC__') < 0:
        conf.fatal('Could not determine the compiler type')
    k = {}
    out = out.splitlines()
    import shlex
    for line in out:
        lst = shlex.split(line)
        if len(lst) > 2:
            key = lst[1]
            val = lst[2]
            k[key] = val

    def isD(var):
        return var in k

    def isT(var):
        return var in k and k[var] != '0'

    conf.env.FC_VERSION = (k['__GNUC__'], k['__GNUC_MINOR__'], k['__GNUC_PATCHLEVEL__'])


def configure(conf):
    conf.find_gfortran()
    conf.find_ar()
    conf.fc_flags()
    conf.fc_add_flags()
    conf.gfortran_flags()
    conf.gfortran_modifier_platform()
    conf.check_gfortran_o_space()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, tempfile, optparse, sys, re
from waflib import Logs, Utils, Context, Errors

options = optparse.Values()
commands = []
envvars = []
lockfile = os.environ.get('WAFLOCK', '.lock-waf_%s_build' % sys.platform)


class opt_parser(optparse.OptionParser):
    def __init__(self, ctx, allow_unknown=False):
        optparse.OptionParser.__init__(
            self,
            conflict_handler='resolve',
            add_help_option=False,
            version='%s %s (%s)' % (Context.WAFNAME, Context.WAFVERSION, Context.WAFREVISION)
        )
        self.formatter.width = Logs.get_term_cols()
        self.ctx = ctx
        self.allow_unknown = allow_unknown

    def _process_args(self, largs, rargs, values):
        while rargs:
            try:
                optparse.OptionParser._process_args(self, largs, rargs, values)
            except (optparse.BadOptionError, optparse.AmbiguousOptionError) as e:
                if self.allow_unknown:
                    largs.append(e.opt_str)
                else:
                    self.error(str(e))

    def _process_long_opt(self, rargs, values):
        if self.allow_unknown:
            back = [] + rargs
            try:
                optparse.OptionParser._process_long_opt(self, rargs, values)
            except optparse.BadOptionError:
                while rargs:
                    rargs.pop()
                rargs.extend(back)
                rargs.pop(0)
                raise
        else:
            optparse.OptionParser._process_long_opt(self, rargs, values)

    def print_usage(self, file=None):
        return self.print_help(file)

    def get_usage(self):
        cmds_str = {}
        for cls in Context.classes:
            if not cls.cmd or cls.cmd == 'options' or cls.cmd.startswith('_'):
                continue
            s = cls.__doc__ or ''
            cmds_str[cls.cmd] = s
        if Context.g_module:
            for (k, v) in Context.g_module.__dict__.items():
                if k in ('options', 'init', 'shutdown'):
                    continue
                if type(v) is type(Context.create_context):
                    if v.__doc__ and not k.startswith('_'):
                        cmds_str[k] = v.__doc__
        just = 0
        for k in cmds_str:
            just = max(just, len(k))
        lst = ['  %s: %s' % (k.ljust(just), v) for (k, v) in cmds_str.items()]
        lst.sort()
        ret = '\n'.join(lst)
        return '''%s [commands] [options]

Main commands (example: ./%s build -j4)
%s
''' % (Context.WAFNAME, Context.WAFNAME, ret)


class OptionsContext(Context.Context):
    cmd = 'options'
    fun = 'options'

    def __init__(self, **kw):
        super(OptionsContext, self).__init__(**kw)
        self.parser = opt_parser(self)
        self.option_groups = {}
        jobs = self.jobs()
        p = self.add_option
        color = os.environ.get('NOCOLOR', '') and 'no' or 'auto'
        if os.environ.get('CLICOLOR', '') == '0':
            color = 'no'
        elif os.environ.get('CLICOLOR_FORCE', '') == '1':
            color = 'yes'
        p(
            '-c',
            '--color',
            dest='colors',
            default=color,
            action='store',
            help='whether to use colors (yes/no/auto) [default: auto]',
            choices=('yes', 'no', 'auto')
        )
        p('-j', '--jobs', dest='jobs', default=jobs, type='int', help='amount of parallel jobs (%r)' % jobs)
        p('-k', '--keep', dest='keep', default=0, action='count', help='continue despite errors (-kk to try harder)')
        p(
            '-v',
            '--verbose',
            dest='verbose',
            default=0,
            action='count',
            help='verbosity level -v -vv or -vvv [default: 0]'
        )
        p('--zones', dest='zones', default='', action='store', help='debugging zones (task_gen, deps, tasks, etc)')
        p('--profile', dest='profile', default=0, action='store_true', help=optparse.SUPPRESS_HELP)
        p('--pdb', dest='pdb', default=0, action='store_true', help=optparse.SUPPRESS_HELP)
        p('-h', '--help', dest='whelp', default=0, action='store_true', help="show this help message and exit")
        gr = self.add_option_group('Configuration options')
        self.option_groups['configure options'] = gr
        gr.add_option('-o', '--out', action='store', default='', help='build dir for the project', dest='out')
        gr.add_option('-t', '--top', action='store', default='', help='src dir for the project', dest='top')
        gr.add_option(
            '--no-lock-in-run',
            action='store_true',
            default=os.environ.get('NO_LOCK_IN_RUN', ''),
            help=optparse.SUPPRESS_HELP,
            dest='no_lock_in_run'
        )
        gr.add_option(
            '--no-lock-in-out',
            action='store_true',
            default=os.environ.get('NO_LOCK_IN_OUT', ''),
            help=optparse.SUPPRESS_HELP,
            dest='no_lock_in_out'
        )
        gr.add_option(
            '--no-lock-in-top',
            action='store_true',
            default=os.environ.get('NO_LOCK_IN_TOP', ''),
            help=optparse.SUPPRESS_HELP,
            dest='no_lock_in_top'
        )
        default_prefix = getattr(Context.g_module, 'default_prefix', os.environ.get('PREFIX'))
        if not default_prefix:
            if Utils.unversioned_sys_platform() == 'win32':
                d = tempfile.gettempdir()
                default_prefix = d[0].upper() + d[1:]
            else:
                default_prefix = '/usr/local/'
        gr.add_option(
            '--prefix',
            dest='prefix',
            default=default_prefix,
            help='installation prefix [default: %r]' % default_prefix
        )
        gr.add_option('--bindir', dest='bindir', help='bindir')
        gr.add_option('--libdir', dest='libdir', help='libdir')
        gr = self.add_option_group('Build and installation options')
        self.option_groups['build and install options'] = gr
        gr.add_option(
            '-p',
            '--progress',
            dest='progress_bar',
            default=0,
            action='count',
            help='-p: progress bar; -pp: ide output'
        )
        gr.add_option(
            '--targets', dest='targets', default='', action='store', help='task generators, e.g. "target1,target2"'
        )
        gr = self.add_option_group('Step options')
        self.option_groups['step options'] = gr
        gr.add_option(
            '--files',
            dest='files',
            default='',
            action='store',
            help='files to process, by regexp, e.g. "*/main.c,*/test/main.o"'
        )
        default_destdir = os.environ.get('DESTDIR', '')
        gr = self.add_option_group('Installation and uninstallation options')
        self.option_groups['install/uninstall options'] = gr
        gr.add_option(
            '--destdir',
            help='installation root [default: %r]' % default_destdir,
            default=default_destdir,
            dest='destdir'
        )
        gr.add_option('-f', '--force', dest='force', default=False, action='store_true', help='force file installation')
        gr.add_option(
            '--distcheck-args', metavar='ARGS', help='arguments to pass to distcheck', default=None, action='store'
        )

    def jobs(self):
        count = int(os.environ.get('JOBS', 0))
        if count < 1:
            if 'NUMBER_OF_PROCESSORS' in os.environ:
                count = int(os.environ.get('NUMBER_OF_PROCESSORS', 1))
            else:
                if hasattr(os, 'sysconf_names'):
                    if 'SC_NPROCESSORS_ONLN' in os.sysconf_names:
                        count = int(os.sysconf('SC_NPROCESSORS_ONLN'))
                    elif 'SC_NPROCESSORS_CONF' in os.sysconf_names:
                        count = int(os.sysconf('SC_NPROCESSORS_CONF'))
                if not count and os.name not in ('nt', 'java'):
                    try:
                        tmp = self.cmd_and_log(['sysctl', '-n', 'hw.ncpu'], quiet=0)
                    except Errors.WafError:
                        pass
                    else:
                        if re.match('^[0-9]+$', tmp):
                            count = int(tmp)
        if count < 1:
            count = 1
        elif count > 1024:
            count = 1024
        return count

    def add_option(self, *k, **kw):
        return self.parser.add_option(*k, **kw)

    def add_option_group(self, *k, **kw):
        try:
            gr = self.option_groups[k[0]]
        except KeyError:
            gr = self.parser.add_option_group(*k, **kw)
        self.option_groups[k[0]] = gr
        return gr

    def get_option_group(self, opt_str):
        try:
            return self.option_groups[opt_str]
        except KeyError:
            for group in self.parser.option_groups:
                if group.title == opt_str:
                    return group
            return None

    def sanitize_path(self, path, cwd=None):
        if not cwd:
            cwd = Context.launch_dir
        p = os.path.expanduser(path)
        p = os.path.join(cwd, p)
        p = os.path.normpath(p)
        p = os.path.abspath(p)
        return p

    def parse_cmd_args(self, _args=None, cwd=None, allow_unknown=False):
        self.parser.allow_unknown = allow_unknown
        (options, leftover_args) = self.parser.parse_args(args=_args)
        envvars = []
        commands = []
        for arg in leftover_args:
            if '=' in arg:
                envvars.append(arg)
            elif arg != 'options':
                commands.append(arg)
        if options.jobs < 1:
            options.jobs = 1
        for name in 'top out destdir prefix bindir libdir'.split():
            if getattr(options, name, None):
                path = self.sanitize_path(getattr(options, name), cwd)
                setattr(options, name, path)
        return options, commands, envvars

    def init_module_vars(self, arg_options, arg_commands, arg_envvars):
        options.__dict__.clear()
        del commands[:]
        del envvars[:]
        options.__dict__.update(arg_options.__dict__)
        commands.extend(arg_commands)
        envvars.extend(arg_envvars)
        for var in envvars:
            (name, value) = var.split('=', 1)
            os.environ[name.strip()] = value

    def init_logs(self, options, commands, envvars):
        Logs.verbose = options.verbose
        if options.verbose >= 1:
            self.load('errcheck')
        colors = {'yes': 2, 'auto': 1, 'no': 0}[options.colors]
        Logs.enable_colors(colors)
        if options.zones:
            Logs.zones = options.zones.split(',')
            if not Logs.verbose:
                Logs.verbose = 1
        elif Logs.verbose > 0:
            Logs.zones = ['runner']
        if Logs.verbose > 2:
            Logs.zones = ['*']

    def parse_args(self, _args=None):
        options, commands, envvars = self.parse_cmd_args()
        self.init_logs(options, commands, envvars)
        self.init_module_vars(options, commands, envvars)

    def execute(self):
        super(OptionsContext, self).execute()
        self.parse_args()
        Utils.alloc_process_pool(options.jobs)



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import pytest

from PyInstaller.compat import is_win, is_linux
from PyInstaller.utils.tests import importorskip, skipif, requires
from PyInstaller.utils.hooks import can_import_module


@importorskip('gevent')
def test_gevent(pyi_builder):
    pyi_builder.test_source("""
        import gevent
        gevent.spawn(lambda: x)
        """)


@importorskip('gevent')
def test_gevent_monkey(pyi_builder):
    pyi_builder.test_source("""
        from gevent.monkey import patch_all
        patch_all()
        """)


# The tkinter module may be available for import, but not actually importable due to missing shared libraries.
# Therefore, we need to use `can_import_module`-based skip decorator instead of `@importorskip`.
@pytest.mark.skipif(not can_import_module("tkinter"), reason="tkinter cannot be imported.")
def test_tkinter(pyi_builder):
    pyi_builder.test_script('pyi_lib_tkinter.py')


# In contrast to test_tkinter, which performs basic import test and verifies that the environment variables are properly
# set, this is a full functional test; we try to create a Tk window with label and button, and register a timer to shut
# down the application. Doing so verifies that all Tcl/Tk files (e.g., .tcl scripts from library directories) are
# properly collected.
#
# The prerequisite for this test is that tkinter can be used unfrozen, so try instantiating a window in a subprocess
# to verify that this is the case. This check should cover the following scenarios:
#  - tkinter missing
#  - import of tkinter crashes python interpreter
#  - tkinter.Tk() fails due to DISPLAY not being set on linux
#  - tkinter.Tk() fails due to faulty build (e.g., due to Tcl/Tk version mix-up, as seen with python <= 3.10 builds on
#    macos-12 GHA runners; https://github.com/actions/setup-python/issues/649#issuecomment-1745056485)
def _tkinter_fully_usable():
    from PyInstaller import isolated

    @isolated.decorate
    def _create_tkinter_window():
        import tkinter
        tkinter.Tk()

    try:
        _create_tkinter_window()
    except Exception:
        return False

    return True


def test_tkinter_functional(pyi_builder):
    if not _tkinter_fully_usable():
        pytest.skip("tkinter is not fully usable.")

    pyi_builder.test_source(
        """
        import tkinter
        import tkinter.messagebox

        root = tkinter.Tk()

        # Dump information about library/data directory.
        tcl_dir = root.tk.exprstring('$tcl_library')
        print(f"Run-time Tcl library/data directory: {tcl_dir}")

        tk_dir = root.tk.exprstring('$tk_library')
        print(f"Run-time Tk library/data directory: {tk_dir}")

        # Create test GUI
        label = tkinter.Label(root, text="Hello World")
        label.pack()

        def test_button_callback():
            tkinter.messagebox.showinfo("Test", "Test message")

        button = tkinter.Button(root, text="Test button", command=test_button_callback)
        button.pack()

        def shutdown_timer_callback():
            print("Shutting down!")
            root.destroy()

        shutdown_interval = 1000  # ms
        print(f"Starting shutdown timer ({shutdown_interval} ms)...")
        root.after(shutdown_interval, shutdown_timer_callback)

        print("Entering main loop...")
        root.mainloop()

        print("Done!")
    """
    )


@importorskip('sphinx')
def test_sphinx(pyi_builder, data_dir):
    pyi_builder.test_script(
        'pyi_lib_sphinx.py',
        # Pass the path to temporary copy of data directory via command-line arguments.
        app_args=[str(data_dir)],
    )


@importorskip('pygments')
def test_pygments(pyi_builder):
    pyi_builder.test_source(
        """
        # This sample code is taken from http://pygments.org/docs/quickstart/.
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.formatters import HtmlFormatter

        code = 'print "Hello World"'
        print(highlight(code, PythonLexer(), HtmlFormatter()))
        """
    )


@requires('zope.interface')
def test_zope_interface(pyi_builder):
    # Tests that `nspkg.pth`-based namespace package are bundled properly. The `nspkg.pth` file is created by
    # setuptools and thus changes frequently. If this test fails, most probably _SETUPTOOLS_NAMESPACEPKG_PTHs
    # in modulegraph needs to be updated.
    pyi_builder.test_source(
        """
        # Package 'zope' does not contain __init__.py file.
        # Just importing 'zope.interface' is sufficient.
        import zope.interface
        """
    )


# The tkinter module may be available for import, but not actually importable due to missing shared libraries.
# Therefore, we need to use `can_import_module`-based skip decorator instead of `@importorskip`.
@importorskip('idlelib')
@pytest.mark.skipif(not can_import_module("tkinter"), reason="tkinter cannot be imported.")
def test_idlelib(pyi_builder):
    pyi_builder.test_source(
        """
        # This file depends on loading some icons, located based on __file__.
        import idlelib.tree
        """
    )


@importorskip('keyring')
@skipif(
    is_linux,
    reason="SecretStorage backend on linux requires active D-BUS session and initialized keyring, and may "
    "need to unlock the keyring via UI prompt."
)
def test_keyring(pyi_builder):
    pyi_builder.test_source("""
        import keyring
        keyring.get_password("test", "test")
        """)


@importorskip('numpy')
def test_numpy(pyi_builder):
    pyi_builder.test_source(
        """
        import numpy
        from numpy.core.numeric import dot
        print('dot(3, 4):', dot(3, 4))
        """
    )


@importorskip('pytz')
def test_pytz(pyi_builder):
    pyi_builder.test_source("""
        import pytz
        pytz.timezone('US/Eastern')
        """)


@importorskip('requests')
def test_requests(pyi_builder, data_dir):
    # NOTE: including the `data_dir` fixture copies files needed by this test!
    # We collect the data into frozen application.
    add_data_arg = f"{data_dir / '*'}:."
    pyi_builder.test_script('pyi_lib_requests.py', pyi_args=['--add-data', add_data_arg])


@importorskip('urllib3.packages.six')
def test_urllib3_six(pyi_builder):
    # Test for pre-safe-import urllib3.packages.six.moves.
    pyi_builder.test_source(
        """
        import urllib3.connectionpool
        import types
        assert isinstance(urllib3.connectionpool.queue, types.ModuleType)
        """
    )


@importorskip('sqlite3')
def test_sqlite3(pyi_builder):
    pyi_builder.test_source(
        """
        # PyInstaller did not included module 'sqlite3.dump'.
        import sqlite3
        conn = sqlite3.connect(':memory:')
        csr = conn.cursor()
        csr.execute('CREATE TABLE Example (id)')
        for line in conn.iterdump():
             print(line)
        """
    )


@requires('scapy >= 2.0')
def test_scapy(pyi_builder):
    pyi_builder.test_source(
        """
        # Test-cases taken from issue #834
        import scapy.all
        scapy.all.IP

        from scapy.all import IP

        # Test-case taken from issue #202.
        from scapy.all import *
        DHCP  # scapy.layers.dhcp.DHCP
        BOOTP  # scapy.layers.dhcp.BOOTP
        DNS  # scapy.layers.dns.DNS
        ICMP  # scapy.layers.inet.ICMP
        """
    )


@requires('scapy >= 2.0')
def test_scapy2(pyi_builder):
    pyi_builder.test_source(
        """
        # Test the hook to scapy.layers.all
        from scapy.layers.all import DHCP
        """
    )


@requires('scapy >= 2.0')
def test_scapy3(pyi_builder):
    pyi_builder.test_source(
        """
        # Test whether
        # a) scapy packet layers are not included if neither scapy.all nor scapy.layers.all are imported
        # b) packages are included if imported explicitly

        NAME = 'hook-scapy.layers.all'
        layer_inet = 'scapy.layers.inet'

        def testit():
            try:
                __import__(layer_inet)
                raise SystemExit('Self-test of hook %s failed: package module found'
                                 % NAME)
            except ImportError, e:
                if not e.args[0].endswith(' inet'):
                    raise SystemExit('Self-test of hook %s failed: package module found and has import errors: %r'
                                     % (NAME, e))

        import scapy
        testit()
        import scapy.layers
        testit()
        # Explicitly import a single layer module. Note: This module MUST NOT import inet (neither directly nor
        # indirectly), otherwise the test above fails.
        import scapy.layers.ir
        """
    )


@importorskip('sqlalchemy')
def test_sqlalchemy(pyi_builder):
    pyi_builder.test_source(
        """
        # The hook behaviour is to include with sqlalchemy all installed database backends.
        import sqlalchemy
        # This import was known to fail with sqlalchemy 0.9.1
        import sqlalchemy.ext.declarative
        """
    )


@importorskip('twisted')
def test_twisted(pyi_builder):
    pyi_builder.test_source(
        """
        # Twisted is an event-driven networking engine.
        #
        # The 'reactor' is object that starts the eventloop.
        # There are different types of platform specific reactors.
        # Platform specific reactor is wrapped into twisted.internet.reactor module.
        from twisted.internet import reactor
        # Applications importing module twisted.internet.reactor might fail with error like:
        #
        #     AttributeError: 'module' object has no attribute 'listenTCP'
        #
        # Ensure default reactor was loaded - it has method 'listenTCP' to start server.
        if not hasattr(reactor, 'listenTCP'):
            raise SystemExit('Twisted reactor not properly initialized.')
        """
    )


@importorskip('pyexcelerate')
def test_pyexcelerate(pyi_builder):
    pyi_builder.test_source(
        """
        # Requires PyExcelerate 0.6.1 or higher
        # Tested on Windows 7 x64 SP1 with CPython 2.7.6
        import pyexcelerate
        """
    )


@importorskip('usb')
@pytest.mark.skipif(is_linux, reason='libusb_exit segfaults on some linuxes')
def test_usb(pyi_builder):
    # See if the usb package is supported on this platform.
    try:
        import usb
        # This will verify that the backend is present; if not, it will skip this test.
        usb.core.find()
    except (ImportError, usb.core.NoBackendError):
        pytest.skip('USB backnd not found.')

    pyi_builder.test_source(
        """
        import usb.core
        # NoBackendError fails the test if no backends are found.
        usb.core.find()
        """
    )


@importorskip('zeep')
def test_zeep(pyi_builder):
    pyi_builder.test_source(
        """
        # Test the hook to zeep
        from zeep import utils
        utils.get_version()
        """
    )


@importorskip('pandas')
def test_pandas_extension(pyi_builder):
    # Tests that the C extension ``pandas._libs.lib`` is properly bundled. Issue #1580.
    # See http://pandas.pydata.org/pandas-docs/stable/whatsnew.html#modules-privacy-has-changed.
    pyi_builder.test_source(
        """
        from pandas._libs.lib import is_float
        assert is_float(1) == 0
        """
    )


@importorskip('pandas')
@importorskip('jinja2')
def test_pandas_io_formats_style(pyi_builder):
    # pandas.io.formats.style requires jinja2 as hiddenimport, as well as collected template file
    # from pandas/io/formats/templates. See #6008 and #6009.
    pyi_builder.test_source("""
        import pandas.io.formats.style
        """)


@importorskip('pandas')
@importorskip('matplotlib')
def test_pandas_plotting_matplotlib(pyi_builder):
    # Test that pandas.plotting works. Starting with pandas 1.3.0, the used pandas.plotting._matplotlib backend module
    # is loaded via importlib.import_module(), and needs a hidden import. See #5994.
    pyi_builder.test_source(
        """
        import matplotlib as mpl
        import pandas as pd

        mpl.use('Agg')  # Use headless Agg backend to avoid dependency on display server.

        series = pd.Series([0, 1, 2, 3], [0, 1, 2, 3])
        series.plot()
        """
    )


@importorskip('win32ctypes')
@pytest.mark.skipif(not is_win, reason='pywin32-ctypes is supported only on Windows')
@pytest.mark.parametrize('submodule', ['win32api', 'win32cred', 'pywintypes'])
def test_pywin32ctypes(pyi_builder, submodule):
    pyi_builder.test_source(f"""
        from win32ctypes.pywin32 import {submodule}
        """)


@importorskip('setuptools')
def test_setuptools(pyi_builder):
    pyi_builder.test_source("""
        import setuptools
        """)


@importorskip('babel')
def test_babel(pyi_builder):
    # Try to format a date/time in order to ensure that data files from babel's locale-data directory (especially
    # root.dat) can be unpickled.
    pyi_builder.test_source(
        """
        import datetime
        from babel.dates import format_datetime

        datetime_obj = datetime.datetime(2007, 4, 1, 15, 30)
        print(format_datetime(datetime_obj, 'full', locale='fr_FR'))
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import struct

import pefile

from PyInstaller.compat import win32api


def pefile_check_control_flow_guard(filename):
    """
    Checks if the specified PE file has CFG (Control Flow Guard) enabled.

    Parameters
    ----------
    filename : str
        Path to the PE file to inspect.

    Returns
    ----------
    bool
        True if file is a PE file with CFG enabled. False if CFG is not enabled or if file could not be processed using
        the pefile library.
    """
    try:
        pe = pefile.PE(filename, fast_load=True)
        # https://docs.microsoft.com/en-us/windows/win32/debug/pe-format
        # IMAGE_DLLCHARACTERISTICS_GUARD_CF = 0x4000
        return bool(pe.OPTIONAL_HEADER.DllCharacteristics & 0x4000)
    except Exception:
        return False


# Ensures no code from the executable is executed.
LOAD_LIBRARY_AS_DATAFILE = 2


def getRaw(text):
    """
    Encodes text as UTF-16LE (Microsoft 'Unicode') for use in structs.
    """
    return text.encode('UTF-16LE')


def read_version_info_from_executable(exe_filename):
    """
    Read the version information structure from the given executable's resources, and return it as an instance of
    `VSVersionInfo` structure.
    """
    h = win32api.LoadLibraryEx(exe_filename, 0, LOAD_LIBRARY_AS_DATAFILE)
    res = win32api.EnumResourceNames(h, pefile.RESOURCE_TYPE['RT_VERSION'])
    if not len(res):
        return None
    data = win32api.LoadResource(h, pefile.RESOURCE_TYPE['RT_VERSION'], res[0])
    info = VSVersionInfo()
    info.fromRaw(data)
    win32api.FreeLibrary(h)
    return info


def nextDWord(offset):
    """
    Align `offset` to the next 4-byte boundary.
    """
    return ((offset + 3) >> 2) << 2


class VSVersionInfo:
    """
    WORD  wLength;        // length of the VS_VERSION_INFO structure
    WORD  wValueLength;   // length of the Value member
    WORD  wType;          // 1 means text, 0 means binary
    WCHAR szKey[];        // Contains the Unicode string "VS_VERSION_INFO".
    WORD  Padding1[];
    VS_FIXEDFILEINFO Value;
    WORD  Padding2[];
    WORD  Children[];     // zero or more StringFileInfo or VarFileInfo
                          // structures (or both) that are children of the
                          // current version structure.
    """
    def __init__(self, ffi=None, kids=None):
        self.ffi = ffi
        self.kids = kids or []

    def fromRaw(self, data):
        i, (sublen, vallen, wType, nm) = parseCommon(data)
        #vallen is length of the ffi, typ is 0, nm is 'VS_VERSION_INFO'.
        i = nextDWord(i)
        # Now a VS_FIXEDFILEINFO
        self.ffi = FixedFileInfo()
        j = self.ffi.fromRaw(data, i)
        i = j
        while i < sublen:
            j = i
            i, (csublen, cvallen, ctyp, nm) = parseCommon(data, i)
            if nm.strip() == 'StringFileInfo':
                sfi = StringFileInfo()
                k = sfi.fromRaw(csublen, cvallen, nm, data, i, j + csublen)
                self.kids.append(sfi)
                i = k
            else:
                vfi = VarFileInfo()
                k = vfi.fromRaw(csublen, cvallen, nm, data, i, j + csublen)
                self.kids.append(vfi)
                i = k
            i = j + csublen
            i = nextDWord(i)
        return i

    def toRaw(self):
        raw_name = getRaw('VS_VERSION_INFO')
        rawffi = self.ffi.toRaw()
        vallen = len(rawffi)
        typ = 0
        sublen = 6 + len(raw_name) + 2
        pad = b''
        if sublen % 4:
            pad = b'\000\000'
        sublen = sublen + len(pad) + vallen
        pad2 = b''
        if sublen % 4:
            pad2 = b'\000\000'
        tmp = b''.join([kid.toRaw() for kid in self.kids])
        sublen = sublen + len(pad2) + len(tmp)
        return struct.pack('HHH', sublen, vallen, typ) + raw_name + b'\000\000' + pad + rawffi + pad2 + tmp

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        indent = indent + '  '
        tmp = [kid.__str__(indent + '  ') for kid in self.kids]
        tmp = ', \n'.join(tmp)
        return '\n'.join([
            "# UTF-8",
            "#",
            "# For more details about fixed file info 'ffi' see:",
            "# http://msdn.microsoft.com/en-us/library/ms646997.aspx",
            "VSVersionInfo(",
            indent + f"ffi={self.ffi.__str__(indent)},",
            indent + "kids=[",
            tmp,
            indent + "]",
            ")",
        ])

    def __repr__(self):
        return "versioninfo.VSVersionInfo(ffi=%r, kids=%r)" % (self.ffi, self.kids)


def parseCommon(data, start=0):
    i = start + 6
    (wLength, wValueLength, wType) = struct.unpack('3H', data[start:i])
    i, text = parseUString(data, i, i + wLength)
    return i, (wLength, wValueLength, wType, text)


def parseUString(data, start, limit):
    i = start
    while i < limit:
        if data[i:i + 2] == b'\000\000':
            break
        i += 2
    text = data[start:i].decode('UTF-16LE')
    i += 2
    return i, text


class FixedFileInfo:
    """
    DWORD dwSignature;        //Contains the value 0xFEEFO4BD
    DWORD dwStrucVersion;     // binary version number of this structure.
                              // The high-order word of this member contains
                              // the major version number, and the low-order
                              // word contains the minor version number.
    DWORD dwFileVersionMS;    // most significant 32 bits of the file's binary
                              // version number
    DWORD dwFileVersionLS;    //
    DWORD dwProductVersionMS; // most significant 32 bits of the binary version
                              // number of the product with which this file was
                              // distributed
    DWORD dwProductVersionLS; //
    DWORD dwFileFlagsMask;    // bitmask that specifies the valid bits in
                              // dwFileFlags. A bit is valid only if it was
                              // defined when the file was created.
    DWORD dwFileFlags;        // VS_FF_DEBUG, VS_FF_PATCHED etc.
    DWORD dwFileOS;           // VOS_NT, VOS_WINDOWS32 etc.
    DWORD dwFileType;         // VFT_APP etc.
    DWORD dwFileSubtype;      // 0 unless VFT_DRV or VFT_FONT or VFT_VXD
    DWORD dwFileDateMS;
    DWORD dwFileDateLS;
    """
    def __init__(
        self,
        filevers=(0, 0, 0, 0),
        prodvers=(0, 0, 0, 0),
        mask=0x3f,
        flags=0x0,
        OS=0x40004,
        fileType=0x1,
        subtype=0x0,
        date=(0, 0)
    ):
        self.sig = 0xfeef04bd
        self.strucVersion = 0x10000
        self.fileVersionMS = (filevers[0] << 16) | (filevers[1] & 0xffff)
        self.fileVersionLS = (filevers[2] << 16) | (filevers[3] & 0xffff)
        self.productVersionMS = (prodvers[0] << 16) | (prodvers[1] & 0xffff)
        self.productVersionLS = (prodvers[2] << 16) | (prodvers[3] & 0xffff)
        self.fileFlagsMask = mask
        self.fileFlags = flags
        self.fileOS = OS
        self.fileType = fileType
        self.fileSubtype = subtype
        self.fileDateMS = date[0]
        self.fileDateLS = date[1]

    def fromRaw(self, data, i):
        (
            self.sig,
            self.strucVersion,
            self.fileVersionMS,
            self.fileVersionLS,
            self.productVersionMS,
            self.productVersionLS,
            self.fileFlagsMask,
            self.fileFlags,
            self.fileOS,
            self.fileType,
            self.fileSubtype,
            self.fileDateMS,
            self.fileDateLS,
        ) = struct.unpack('13L', data[i:i + 52])
        return i + 52

    def toRaw(self):
        return struct.pack(
            '13L',
            self.sig,
            self.strucVersion,
            self.fileVersionMS,
            self.fileVersionLS,
            self.productVersionMS,
            self.productVersionLS,
            self.fileFlagsMask,
            self.fileFlags,
            self.fileOS,
            self.fileType,
            self.fileSubtype,
            self.fileDateMS,
            self.fileDateLS,
        )

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        fv = (
            self.fileVersionMS >> 16, self.fileVersionMS & 0xffff,
            self.fileVersionLS >> 16, self.fileVersionLS & 0xffff,
        )  # yapf: disable
        pv = (
            self.productVersionMS >> 16, self.productVersionMS & 0xffff,
            self.productVersionLS >> 16, self.productVersionLS & 0xffff,
        )  # yapf: disable
        fd = (self.fileDateMS, self.fileDateLS)
        tmp = [
            'FixedFileInfo(',
            '# filevers and prodvers should be always a tuple with four items: (1, 2, 3, 4)',
            '# Set not needed items to zero 0.',
            'filevers=%s,' % (fv,),
            'prodvers=%s,' % (pv,),
            "# Contains a bitmask that specifies the valid bits 'flags'r",
            'mask=%s,' % hex(self.fileFlagsMask),
            '# Contains a bitmask that specifies the Boolean attributes of the file.',
            'flags=%s,' % hex(self.fileFlags),
            '# The operating system for which this file was designed.',
            '# 0x4 - NT and there is no need to change it.',
            'OS=%s,' % hex(self.fileOS),
            '# The general type of file.',
            '# 0x1 - the file is an application.',
            'fileType=%s,' % hex(self.fileType),
            '# The function of the file.',
            '# 0x0 - the function is not defined for this fileType',
            'subtype=%s,' % hex(self.fileSubtype),
            '# Creation date and time stamp.',
            'date=%s' % (fd,),
            ')',
        ]
        return f'\n{indent}  '.join(tmp)

    def __repr__(self):
        fv = (
            self.fileVersionMS >> 16, self.fileVersionMS & 0xffff,
            self.fileVersionLS >> 16, self.fileVersionLS & 0xffff,
        )  # yapf: disable
        pv = (
            self.productVersionMS >> 16, self.productVersionMS & 0xffff,
            self.productVersionLS >> 16, self.productVersionLS & 0xffff,
        )  # yapf: disable
        fd = (self.fileDateMS, self.fileDateLS)
        return (
            'versioninfo.FixedFileInfo(filevers=%r, prodvers=%r, '
            'mask=0x%x, flags=0x%x, OS=0x%x, '
            'fileType=%r, subtype=0x%x, date=%r)' %
            (fv, pv, self.fileFlagsMask, self.fileFlags, self.fileOS, self.fileType, self.fileSubtype, fd)
        )


class StringFileInfo:
    """
    WORD        wLength;      // length of the version resource
    WORD        wValueLength; // length of the Value member in the current
                              // VS_VERSION_INFO structure
    WORD        wType;        // 1 means text, 0 means binary
    WCHAR       szKey[];      // Contains the Unicode string 'StringFileInfo'.
    WORD        Padding[];
    StringTable Children[];   // list of zero or more String structures
    """
    def __init__(self, kids=None):
        self.name = 'StringFileInfo'
        self.kids = kids or []

    def fromRaw(self, sublen, vallen, name, data, i, limit):
        self.name = name
        while i < limit:
            st = StringTable()
            j = st.fromRaw(data, i, limit)
            self.kids.append(st)
            i = j
        return i

    def toRaw(self):
        raw_name = getRaw(self.name)
        vallen = 0
        typ = 1
        sublen = 6 + len(raw_name) + 2
        pad = b''
        if sublen % 4:
            pad = b'\000\000'
        tmp = b''.join([kid.toRaw() for kid in self.kids])
        sublen = sublen + len(pad) + len(tmp)
        return struct.pack('HHH', sublen, vallen, typ) + raw_name + b'\000\000' + pad + tmp

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        new_indent = indent + '  '
        tmp = ', \n'.join(kid.__str__(new_indent) for kid in self.kids)
        return f'{indent}StringFileInfo(\n{new_indent}[\n{tmp}\n{new_indent}])'

    def __repr__(self):
        return 'versioninfo.StringFileInfo(%r)' % self.kids


class StringTable:
    """
    WORD   wLength;
    WORD   wValueLength;
    WORD   wType;
    WCHAR  szKey[];
    String Children[];    // list of zero or more String structures.
    """
    def __init__(self, name=None, kids=None):
        self.name = name or ''
        self.kids = kids or []

    def fromRaw(self, data, i, limit):
        i, (cpsublen, cpwValueLength, cpwType, self.name) = parseCodePage(data, i, limit)  # should be code page junk
        i = nextDWord(i)
        while i < limit:
            ss = StringStruct()
            j = ss.fromRaw(data, i, limit)
            i = j
            self.kids.append(ss)
            i = nextDWord(i)
        return i

    def toRaw(self):
        raw_name = getRaw(self.name)
        vallen = 0
        typ = 1
        sublen = 6 + len(raw_name) + 2
        tmp = []
        for kid in self.kids:
            raw = kid.toRaw()
            if len(raw) % 4:
                raw = raw + b'\000\000'
            tmp.append(raw)
        tmp = b''.join(tmp)
        sublen += len(tmp)
        return struct.pack('HHH', sublen, vallen, typ) + raw_name + b'\000\000' + tmp

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        new_indent = indent + '  '
        tmp = (',\n' + new_indent).join(str(kid) for kid in self.kids)
        return f"{indent}StringTable(\n{new_indent}'{self.name}',\n{new_indent}[{tmp}])"

    def __repr__(self):
        return 'versioninfo.StringTable(%r, %r)' % (self.name, self.kids)


class StringStruct:
    """
    WORD   wLength;
    WORD   wValueLength;
    WORD   wType;
    WCHAR  szKey[];
    WORD   Padding[];
    String Value[];
    """
    def __init__(self, name=None, val=None):
        self.name = name or ''
        self.val = val or ''

    def fromRaw(self, data, i, limit):
        i, (sublen, vallen, typ, self.name) = parseCommon(data, i)
        limit = i + sublen
        i = nextDWord(i)
        i, self.val = parseUString(data, i, limit)
        return i

    def toRaw(self):
        raw_name = getRaw(self.name)
        raw_val = getRaw(self.val)
        # TODO: document the size of vallen and sublen.
        vallen = len(self.val) + 1  # Number of (wide-)characters, not bytes!
        typ = 1
        sublen = 6 + len(raw_name) + 2
        pad = b''
        if sublen % 4:
            pad = b'\000\000'
        sublen = sublen + len(pad) + (vallen * 2)
        return struct.pack('HHH', sublen, vallen, typ) + raw_name + b'\000\000' + pad + raw_val + b'\000\000'

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        return "StringStruct(%r, %r)" % (self.name, self.val)

    def __repr__(self):
        return 'versioninfo.StringStruct(%r, %r)' % (self.name, self.val)


def parseCodePage(data, i, limit):
    i, (sublen, wValueLength, wType, nm) = parseCommon(data, i)
    return i, (sublen, wValueLength, wType, nm)


class VarFileInfo:
    """
    WORD  wLength;        // length of the version resource
    WORD  wValueLength;   // length of the Value member in the current
                          // VS_VERSION_INFO structure
    WORD  wType;          // 1 means text, 0 means binary
    WCHAR szKey[];        // Contains the Unicode string 'VarFileInfo'.
    WORD  Padding[];
    Var   Children[];     // list of zero or more Var structures
    """
    def __init__(self, kids=None):
        self.kids = kids or []

    def fromRaw(self, sublen, vallen, name, data, i, limit):
        self.sublen = sublen
        self.vallen = vallen
        self.name = name
        i = nextDWord(i)
        while i < limit:
            vs = VarStruct()
            j = vs.fromRaw(data, i, limit)
            self.kids.append(vs)
            i = j
        return i

    def toRaw(self):
        self.vallen = 0
        self.wType = 1
        self.name = 'VarFileInfo'
        raw_name = getRaw(self.name)
        sublen = 6 + len(raw_name) + 2
        pad = b''
        if sublen % 4:
            pad = b'\000\000'
        tmp = b''.join([kid.toRaw() for kid in self.kids])
        self.sublen = sublen + len(pad) + len(tmp)
        return struct.pack('HHH', self.sublen, self.vallen, self.wType) + raw_name + b'\000\000' + pad + tmp

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        return indent + "VarFileInfo([%s])" % ', '.join(str(kid) for kid in self.kids)

    def __repr__(self):
        return 'versioninfo.VarFileInfo(%r)' % self.kids


class VarStruct:
    """
    WORD  wLength;        // length of the version resource
    WORD  wValueLength;   // length of the Value member in the current
                          // VS_VERSION_INFO structure
    WORD  wType;          // 1 means text, 0 means binary
    WCHAR szKey[];        // Contains the Unicode string 'Translation'
                          // or a user-defined key string value
    WORD  Padding[];      //
    WORD  Value[];        // list of one or more values that are language
                          // and code-page identifiers
    """
    def __init__(self, name=None, kids=None):
        self.name = name or ''
        self.kids = kids or []

    def fromRaw(self, data, i, limit):
        i, (self.sublen, self.wValueLength, self.wType, self.name) = parseCommon(data, i)
        i = nextDWord(i)
        for j in range(0, self.wValueLength, 2):
            kid = struct.unpack('H', data[i:i + 2])[0]
            self.kids.append(kid)
            i += 2
        return i

    def toRaw(self):
        self.wValueLength = len(self.kids) * 2
        self.wType = 0
        raw_name = getRaw(self.name)
        sublen = 6 + len(raw_name) + 2
        pad = b''
        if sublen % 4:
            pad = b'\000\000'
        self.sublen = sublen + len(pad) + self.wValueLength
        tmp = b''.join([struct.pack('H', kid) for kid in self.kids])
        return struct.pack('HHH', self.sublen, self.wValueLength, self.wType) + raw_name + b'\000\000' + pad + tmp

    def __eq__(self, other):
        return self.toRaw() == other

    def __str__(self, indent=''):
        return "VarStruct('%s', %r)" % (self.name, self.kids)

    def __repr__(self):
        return 'versioninfo.VarStruct(%r, %r)' % (self.name, self.kids)


def load_version_info_from_text_file(filename):
    """
    Load the `VSVersionInfo` structure from its string-based (`VSVersionInfo.__str__`) serialization by reading the
    text from the file and running it through `eval()`.
    """

    # Read and parse the version file. It may have a byte order marker or encoding cookie - respect it if it does.
    import PyInstaller.utils.misc as miscutils
    with open(filename, 'rb') as fp:
        text = miscutils.decode(fp.read())

    # Deserialize via eval()
    try:
        info = eval(text)
    except Exception as e:
        raise ValueError("Failed to deserialize VSVersionInfo from text-based representation!") from e

    # Sanity check
    assert isinstance(info, VSVersionInfo), \
        f"Loaded incompatible structure type! Expected VSVersionInfo, got: {type(info)!r}"

    return info


def write_version_info_to_executable(exe_filename, info):
    assert isinstance(info, VSVersionInfo)

    # Remember overlay
    pe = pefile.PE(exe_filename, fast_load=True)
    overlay_before = pe.get_overlay()
    pe.close()

    hdst = win32api.BeginUpdateResource(exe_filename, 0)
    win32api.UpdateResource(hdst, pefile.RESOURCE_TYPE['RT_VERSION'], 1, info.toRaw())
    win32api.EndUpdateResource(hdst, 0)

    if overlay_before:
        # Check if the overlay is still present
        pe = pefile.PE(exe_filename, fast_load=True)
        overlay_after = pe.get_overlay()
        pe.close()

        # If the update removed the overlay data, re-append it
        if not overlay_after:
            with open(exe_filename, 'ab') as exef:
                exef.write(overlay_before)



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This little sample application generates a plugin on the fly, and then tries to import it.

import os
import sys

# We first import a static plugin; the application might have certain plugins that it always loads.
try:
    print('Attempting to import static_plugin...')
    mdl = __import__('static_plugin')
except ImportError:
    raise SystemExit('Failed to import the static plugin.')

plugin_contents = """
print('DYNAMIC PLUGIN IMPORTED.')
print('This is some user-generated plugin that does not exist until')
print('  the application starts and other modules in the directory')
print('  are imported (like the static_plugin).')
"""

# Create the dynamic plugin in the same directory as the executable.
if hasattr(sys, 'frozen'):
    program_dir = os.path.abspath(sys.prefix)
else:
    program_dir = os.path.dirname(os.path.abspath(__file__))
plugin_filename = os.path.join(program_dir, 'dynamic_plugin.py')
fp = open(plugin_filename, 'w')
fp.write(plugin_contents)
fp.close()

# Try import dynamic plugin.
is_error = False
try:
    print('Attempting to import dynamic_plugin...')
    mdl = __import__('dynamic_plugin')
except ImportError:
    is_error = True

# Clean up. Remove files dynamic_plugin.py[c]
for f in (plugin_filename, plugin_filename + 'c'):
    try:
        os.remove(plugin_filename)
    except OSError:
        pass

if is_error:
    # Raise exception.
    raise SystemExit('Failed to import the dynamic plugin.')
